import os
from functools import partial
from collections import OrderedDict

from PySide import QtGui, QtCore, QtUiTools
import yaml

from ..core.operations.Operation import Operation
from . import qttools
from . import widgets 
from ..core import pawstools
from ..core import plugins as pgns
from ..core.models.ListModel import ListModel
from .widgets import WorkflowGraphView

class UiManager(QtCore.QObject):
    """
    Uses the QPawsAPI and PySide Qt to provide a widget that controls PAWS.
    """

    def __init__(self,qpaw):
        super(UiManager,self).__init__()
        ui_file = QtCore.QFile(os.path.join(pawstools.sourcedir,'qt','qtui','main.ui'))
        ui_file.open(QtCore.QFile.ReadOnly)
        self.ui = QtUiTools.QUiLoader().load(ui_file)
        ui_file.close()
        self.ui.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        qpaw.set_logmethod( self.msg_board_log )
        qpaw._wf_manager.emitMessage.connect( self.logMessage )
        self.paw = qpaw
        self.make_viewer()
        self.build()
        self.threads = QtCore.QThreadPool()

    @QtCore.Slot(str)
    def logMessage(self,msg):
        self.msg_board_log(msg)

    def msg_board_log(self,msg):
        """Print timestamped message to msg board"""
        #if (self.ui.message_board.verticalScrollBar().value() == 
        #self.ui.message_board.verticalScrollBar().maximum()):
        #    advance_scrollbar = True
        #else:
        #    advance_scrollbar = False
        self.ui.message_board.appendPlainText(
        '- ' + pawstools.timestr() + ': ' + msg)#+ '\n') 
        #if advance_scrollbar:
        self.ui.message_board.verticalScrollBar().setValue(
        self.ui.message_board.verticalScrollBar().maximum())

    def make_viewer(self):
        """
        Set up the tab viewer widget 
        and display the paws logo 
        in the main viewer
        """
        self.ui.viewer_tabwidget.clear()
        self.viewer_widgets = {}

        self._viewer_frame = QtGui.QFrame()
        self._viewer_layout = QtGui.QGridLayout()
        self._viewer_frame.setLayout(self._viewer_layout)
        self.viewer_widgets['main viewer'] = self._viewer_frame 
        self.ui.viewer_tabwidget.addTab(self._viewer_frame,'main viewer')

        # logo scene:
        img_file = os.path.join(pawstools.rootdir,'graphics','paws_icon_white.png')
        pixmap = QtGui.QPixmap(img_file)
        pixmap_item = QtGui.QGraphicsPixmapItem(pixmap)
        scene = QtGui.QGraphicsScene()
        scene.addItem(pixmap_item)
        qwhite = QtGui.QColor(255,255,255,255)
        textitem = scene.addText("v{}".format(pawstools.version))
        textitem.setPos(100,35)
        textitem.setDefaultTextColor(qwhite)
        logo_view = QtGui.QGraphicsView()
        logo_view.setScene(scene)
        self._viewer_layout.addWidget(logo_view,0,0)

        self.ui.setWindowTitle("paws v{}".format(pawstools.version))
        self.ui.setWindowIcon(pixmap)

    def build(self):
        """
        Set up QObjects and model views 
        for communicating with paws objects 
        """
        self.ui.app_control_box.setTitle('Application Controls')
        self.ui.save_state_button.setText("&Save")
        self.ui.save_state_button.clicked.connect(self.save_state)
        self.ui.load_state_button.setText("&Load")
        self.ui.load_state_button.clicked.connect(self.load_state)

        # Workflows ui stuff
        self.paw._wf_manager.wfStopped.connect(self.update_run_wf_button)
        self.ui.workflows_box.setTitle('Workflows')
        if qttools.have_qt47:
            self.ui.wf_name_entry.setPlaceholderText('(enter a name)')
        else:
            self.ui.wf_name_entry.setText('(enter a name)')
        self.ui.wf_name_header.setText('New workflow:')
        self.ui.wf_name_header.setReadOnly(True)
        self.ui.wf_name_header.setAlignment(QtCore.Qt.AlignRight)
        self.ui.wf_name_header.setStyleSheet( "QLineEdit { background-color: transparent }"
        + self.ui.wf_name_header.styleSheet() )
        self.ui.wf_name_header.setStyleSheet( "QLineEdit { border: none }" 
        + self.ui.wf_name_header.styleSheet() )
        # workflow name entry size policies
        self.ui.wf_name_header.setSizePolicy(QtGui.QSizePolicy.Maximum,
        self.ui.wf_name_header.sizePolicy().verticalPolicy())
        self.ui.wf_name_entry.setSizePolicy(QtGui.QSizePolicy.Minimum,
        self.ui.wf_name_entry.sizePolicy().verticalPolicy())
        self.ui.add_workflow_button.setText('Add')
        self.ui.add_workflow_button.clicked.connect( self.add_wf )
        self.paw._wf_manager.wfAdded.connect(self.append_to_wf_selector)
        lm = ListModel(self.paw.list_wf_tags())
        self.ui.wf_selector.setModel(lm)
        self.ui.wf_selector.currentIndexChanged.connect( partial(self.set_wf) )
        self.ui.run_wf_button.setText("&Run")
        self.ui.run_wf_button.clicked.connect(self.toggle_run_wf)
        #self.qwfman.wfdone.connect(self.update_run_wf_button)
        # If for some reason the QPawsAPI (self.paw) calls select_wf(),
        # the UI should also select that wf.
        self.paw.wfSelectionChanged.connect(self.select_wf)

        # Plugins ui stuff
        self.ui.plugins_box.setTitle('Plugins')
        self.ui.plugin_tree.setModel(self.paw._plugin_manager)
        self.ui.plugin_tree.clicked.connect( self.display_plugin_item )
        self.ui.plugin_tree.setRootIndex(self.paw._plugin_manager.root_index())
        if qttools.have_qt47:
            self.ui.plugin_name_entry.setPlaceholderText('(enter a name)')
        else:
            self.ui.plugin_name_entry.setText('(enter a name)')
        self.ui.add_plugin_header.setText('New plugin:')
        self.ui.add_plugin_header.setReadOnly(True)
        self.ui.add_plugin_header.setAlignment(QtCore.Qt.AlignRight)
        self.ui.add_plugin_header.setStyleSheet( "QLineEdit { background-color: transparent }"
        + self.ui.add_plugin_header.styleSheet() )
        self.ui.add_plugin_header.setStyleSheet( "QLineEdit { border: none }" 
        + self.ui.add_plugin_header.styleSheet() )
        # plugin name entry size policies
        self.ui.add_plugin_header.setSizePolicy(QtGui.QSizePolicy.Maximum,
        self.ui.add_plugin_header.sizePolicy().verticalPolicy())
        self.ui.plugin_name_entry.setSizePolicy(QtGui.QSizePolicy.Maximum,
        self.ui.plugin_name_entry.sizePolicy().verticalPolicy())
        self.ui.plugin_selector.setMinimumWidth(160)
        self.ui.plugin_selector.setSizePolicy(QtGui.QSizePolicy.MinimumExpanding,
        self.ui.plugin_selector.sizePolicy().verticalPolicy())
        pgin_lm = ListModel(pgns.plugin_name_list)
        self.ui.plugin_selector.setModel(pgin_lm)
        self.ui.add_plugin_button.setText('Add')
        self.ui.plugin_tree.hideColumn(1)
        self.ui.plugin_tree.hideColumn(2)
        self.ui.plugin_tree.setColumnWidth(0,200)
        self.ui.add_plugin_button.clicked.connect( self.add_plugin )

        self.ui.op_tree.setModel(self.paw._op_manager)
        self.ui.op_tree.setRootIndex(self.paw._op_manager.root_index())
        self.ui.op_tree.clicked.connect( self.display_op_item )
        self.ui.wf_tree.clicked.connect( self.display_wf_item )

        self.ui.message_board.setReadOnly(True)
        self.ui.message_board.insertPlainText('--- MESSAGE BOARD ---') 
        self.ui.op_info_box.insertPlainText('Operation info: click an Operation above to see its documentation here')
        self.msg_board_log('paws is ready') 
        self.ui.op_tree.setColumnWidth(0,200)
        # UI splitter sizing
        self.ui.hsplitter.setStretchFactor(0,3)    
        self.ui.hsplitter.setStretchFactor(1,5)    
        self.ui.hsplitter.setStretchFactor(2,2)    
        self.ui.left_vsplitter.setStretchFactor(0,1)    
        self.ui.left_vsplitter.setStretchFactor(1,3)    
        self.ui.left_vsplitter.setStretchFactor(2,3)    
        self.ui.center_vsplitter.setStretchFactor(0,1)    

    def append_to_wf_selector(self,new_wfname):
        self.ui.wf_selector.model().append_item(new_wfname)

    def select_wf(self,wfname):
        wf_idx = self.ui.wf_selector.model().list_data().index(wfname)
        self.ui.wf_selector.setCurrentIndex(wf_idx)
        self.set_wf_treeview(wfname)

    def set_wf_treeview(self,wfname):
        wf = self.paw.get_wf(wfname)
        self.ui.wf_tree.setModel(wf)
        self.ui.wf_tree.setRootIndex(wf.root_index())
        # hide the selection flag column.
        self.ui.wf_tree.hideColumn(1)
        self.ui.wf_tree.setColumnWidth(0,200)
        self.update_run_wf_button()

    def set_wf(self,wf_selector_idx):
        wfname = self.ui.wf_selector.model().list_data()[wf_selector_idx]
        if not wfname == self.paw.current_wf_name():
            self.paw.select_wf(wfname)
            self.set_wf_treeview(wfname)

    def add_plugin(self):
        pgin_name = self.ui.plugin_name_entry.text()
        if pgin_name in self.paw.list_plugin_tags():
            msg_ui = qttools.message_ui(self.ui)
            msg_ui.setWindowTitle("Plugin Name Error")
            msg = '[{}] Name {} already assigned to a Plugin. '.format(
                __name__,pgin_name) + 'Loaded plugins: {}'.format(
                self.paw.list_plugin_tags())
            msg_ui.message_box.setPlainText(msg)
            msg_ui.show()
        else:
            try:
                idx = self.ui.plugin_selector.currentIndex()
                pgin_uri = self.ui.plugin_selector.model().list_data()[idx]
                pgin = self.paw.load_plugin(pgin_uri)
            except pawstools.PluginLoadError as ex:
                msg_ui = qttools.message_ui(self.ui)
                msg_ui.setWindowTitle("Plugin Load Error")
                msg_ui.message_box.setPlainText(ex.message)
                msg_ui.show()
                return
            try:
                self.paw.add_plugin(pgin_name,pgin)
            except pawstools.PluginNameError as ex:
                msg_ui = qttools.message_ui(self.ui)
                msg_ui.setWindowTitle("Plugin Name Error")
                msg_ui.message_box.setPlainText(ex.message)
                msg_ui.show()
                return
            self.ui.plugin_name_entry.clear()

    def add_wf(self):
        """
        Method for adding workflows through the main UI.
        For this case, the workflow name is inspected
        to ensure that it doesn't clobber an existing workflow.
        """
        wfname = self.ui.wf_name_entry.text()
        if wfname in self.paw.list_wf_tags():
            msg_ui = qttools.message_ui(self.ui)
            msg_ui.setWindowTitle("Workflow Name Error")
            msg = '[{}] Name {} already assigned to a Workflow. '.format(
                __name__,wfname) + 'Loaded workflows: {}'.format(
                self.paw.list_wf_tags())
            msg_ui.message_box.setPlainText(msg)
            msg_ui.show()
        else:
            try:
                self.paw.add_wf(wfname)
            except pawstools.WfNameError as ex:
                msg_ui = qttools.message_ui(self.ui)
                msg_ui.setWindowTitle("Workflow Name Error")
                msg_ui.message_box.setPlainText(ex.message)
                msg_ui.show()
                return
            self.append_to_wf_selector(wfname)
            #self.ui.wf_selector.setCurrentIndex(self.paw.n_wf()-1)
            self.paw.set_wf(wfname)
            self.ui.wf_name_entry.clear()
            # Add a tab to the viewer for this workflow 
            self.add_wf_tab(wfname)

    def add_wf_tab(self,wfname):
        wf = self.paw.get_wf(wfname)
        g = WorkflowGraphView(wf,self.ui)
        tab_idx = self.ui.viewer_tabwidget.addTab(g,wfname) 

    def display_op_item(self,idx):
        """
        Display selected item from the op tree in viewer layout 
        """
        if idx.isValid(): 
            # Change the current selection to this index
            self.ui.op_tree.setCurrentIndex(idx.sibling(idx.row(),0))
            itm_data = self.paw.get_op_from_index(idx)
            itm_uri = self.paw.get_op_uri_from_index(idx)
            op_widg = None
            if itm_data is None:
                # this should be the condition for a not-enabled Operation
                t = str('selected Operation: {} \n'.format(itm_uri)
                + 'This Operation is not currently enabled. '
                + 'Click the "enable" box to enable it. ')
            elif isinstance(itm_data,dict):
                # this should be the condition for a category
                t = str('selected category: {} \n\n'.format(itm_uri)
                + '{}: '.format(itm_uri)) 
                t = t + self.paw._op_manager.print_cat(itm_uri) 
            else:
                # the only remaining case is an enabled Operation
                op = itm_data()
                t = op.description()
                op_widg = widgets.make_widget(op)
            w = QtGui.QTextEdit()
            w.setText(t)
            self.ui.op_info_box.setText(t)

    def display_plugin_item(self,idx):
        """
        Display selected item from the plugin tree in viewer layout 
        """
        if idx.isValid(): 
            itm_data = self.paw.get_plugin_from_index(idx)
            widg = widgets.make_widget(itm_data)
            self.main_display(widg)

    def display_wf_item(self,idx):
        """
        Display selected item from the workflow tree in viewer layout 
        """
        if idx.isValid(): 
            itm_data = self.paw.current_wf().get_data_from_index(idx)
            widg = widgets.make_widget(itm_data)
            self.main_display(widg)

    def main_display(self,widg=None):
        # Bring up the main viewer from the viewer_tabwidget
        self.ui.viewer_tabwidget.setCurrentWidget(self._viewer_frame)
        # Loop through the viewer layout, last to first, clear the frame
        n_widgets = self._viewer_layout.count()
        for i in range(n_widgets-1,-1,-1):
            # QLayout.takeAt returns a LayoutItem
            itm = self._viewer_layout.takeAt(i)
            # get the QWidget of that LayoutItem and close it 
            itm.widget().close()
            del itm
        if widg is not None:
            self._viewer_layout.addWidget(widg,0,0) 

    def toggle_run_wf(self,wfname=None):
        if wfname is None:
            wfname = self.paw.current_wf_name()
        if wfname is not None:
            if self.paw.is_wf_running(wfname):
                self.stop_wf(wfname)
            else:
                self.start_wf(wfname)

    def start_wf(self,wfname):
        self.ui.run_wf_button.setText("S&top")
        # NOTE: this is where we choose between threaded or not.
        # TODO: consider exposing both run modes to the user.
        self.paw.run_wf(wfname,self.threads)
        #self.paw.run_wf(wfname,None)

    def stop_wf(self,wfname):
        # TODO: the interruption will have to be carried out by signals and slots.
        # the signal should be emitted in QPawsAPI.stop_wf and the slot should be in QWorkflow.
        # TODO: consider how to interrupt a running operation,
        # e.g. a batch execution controller.
        self.paw.stop_wf(wfname)
        self.ui.run_wf_button.setText("&Run")

    def update_run_wf_button(self,wfname=None):
        """
        If the input wfname indicates the currently selected workflow,
        make the self.ui.run_wf_button sane wrt this workflow's status.
        """
        if self.paw.current_wf_name() == wfname:
            if self.paw.is_wf_running(wfname):
                self.ui.run_wf_button.setText("S&top")
            else:
                self.ui.run_wf_button.setText("&Run")

    def save_state(self):
        """
        Start a modal window dialog to choose a .wfl to save the current configuration  
        """
        save_ui = qttools.start_save_ui(self.ui)
        save_ui.setWindowTitle('paws saver')
        #wf_idx = self.ui.wf_selector.currentIndex()
        #wfname = self.ui.wf_selector.model().list_data()[wf_idx]
        save_ui.tree_box.setTitle('Select or enter a .wfl file to save current configuration.')
        save_ui.save_button.clicked.connect( partial(self.finish_save_state,save_ui) )
        save_ui.show()
        save_ui.activateWindow()

    def load_state(self):
        """
        Start a modal window dialog to choose a .wfl to load a previously saved configuration 
        """
        load_ui = qttools.start_load_ui(self.ui)
        load_ui.setWindowTitle('paws loader')
        load_ui.tree_box.setTitle('Select a .wfl file to load a saved paws configuration.')
        load_ui.load_button.clicked.connect( partial(self.finish_load_state,load_ui) )
        load_ui.show()
        load_ui.activateWindow()
      
    def finish_save_state(self,ui):
        fname = ui.filename.text()
        self.paw.save_to_wfl(fname)
        ui.close()

    def finish_load_state(self,ui):
        fname = ui.tree.model().filePath(ui.tree.currentIndex())
        self.paw.load_from_wfl(fname)
        ui.close()




from __future__ import print_function
from collections import OrderedDict
import importlib

from ..models.TreeModel import TreeModel
from ..models.TreeItem import TreeItem
from .. import operations as ops
from ..operations.Operation import Operation

class OpManager(TreeModel):
    """
    Tree structure for categorized storage and retrieval of Operations.
    """

    def __init__(self):
        default_flags = OrderedDict()
        default_flags['enable'] = False
        super(OpManager,self).__init__(default_flags)
        self.logmethod = print 
        self._n_ops = 0
        #self._n_cats = 0

    # override TreeModel.create_tree_item() to add 
    # the correct enable/disable flags. 
    def create_tree_item(self,parent_itm,itm_tag):
        itm = TreeItem(parent_itm,itm_tag)
        op_uri = self.build_uri(itm)
        itm.flags['enable'] = False
        if op_uri in ops.load_flags.keys():
            itm.flags['enable'] = ops.load_flags[op_uri]
        return itm

    def n_ops(self):
        return self._n_ops

    def list_ops(self):
        return [catnm+'.'+opnm for catnm,opnm in ops.cat_op_list] 

    def load_cats(self,cat_list):
        for cat in cat_list:
            itm = self._root_item
            cat_tags = cat.split('.')
            cat_uri = cat_tags[0]
            if not self.contains_uri(cat_uri):
                self.set_item(cat_uri,{})
                #self._n_cats += 1
            if len(cat_tags) > 1:
                for cat_tag in cat_tags[1:]:
                    #if not cat_tag in self.list_child_tags(cat_uri):
                    cat_uri = cat_uri+'.'+cat_tag
                    if not self.contains_uri(cat_uri):
                        #print('set cat {}'.format(cat_uri))
                        self.set_item(cat_uri,{})
                        #self._n_cats += 1

    def load_ops(self,cat_op_list):
        """
        Load OpManager tree from input cat_op_list.
        Format of cat_op_list is [(category1,opname1),(category2,opname2),...].
        i.e. each operation in cat_op_list is specified by a tuple,
        where the first element is a category,
        and the second element is the name of the Operation.
        load_cats() should be called before load_ops()
        and should ensure that all cats in cat_op_list exist in the tree.
        """
        for cat_op in cat_op_list:
            self.add_op(cat_op[0],cat_op[1])
            self._n_ops += 1

    def add_op(self,cat,opname):
        """
        Add op name to the tree under category cat.
        If ops.load_flags indicates that this op should be enabled,
        try to enable it (this causes it to import the module).
        """
        op_uri = cat+'.'+opname
        if ops.load_flags[op_uri]:
            try: 
                self.set_op_enabled(op_uri)
            except ImportError:
                self.logmethod('import error for {}: disabling operation'.format(op_uri))
        else:
            self.set_item(op_uri,None)

    def set_op_enabled(self,op_uri,flag=True):
        cat = op_uri[:op_uri.rfind('.')]
        opname = op_uri.split('.')[-1]
        if flag:
            mod = importlib.import_module('.'+op_uri,ops.__name__)
            op = getattr(mod,opname)
            optest = op()
            self.set_item(op_uri,op)
            ops.load_flags[op_uri] = True
        else:
            # disable the op: set ops.load_flags so that
            # add_op() replaces the treedata with None
            ops.load_flags[op_uri] = False
            self.set_item(op_uri,None)

    def remove_op(self,op_uri):
        """Remove op from the tree by its full category.opname uri"""
        self.remove_item(op_uri)
        self._n_ops -= 1

    def print_cat(self,cat_uri,rowprefix='    '):
        """
        Generate a string that lists the contents 
        of the operations category specified by cat_uri
        """
        catdata = self.get_data_from_uri(cat_uri)
        tree_string = '\n' 
        if isinstance(catdata,dict):
            for k,x in catdata.items():
                if x is None:
                    # this should be the case for not-enabled ops
                    tree_string = tree_string + rowprefix + '{} (disabled) \n'.format(k)
                elif isinstance(x,dict):
                    # this should be the case for a subcat
                    next_cat_tree = self.print_cat(cat_uri+'.'+k,rowprefix+'    ')
                    tree_string = tree_string + rowprefix + '{}: {}'.format(k,next_cat_tree)
                else:
                    # the only remaining case is an enabled operation
                    tree_string = tree_string + rowprefix + '{} \n'.format(k)
        return tree_string






from __future__ import print_function
from collections import OrderedDict
import copy

import numpy as np
from scipy.optimize import minimize as scipimin

def compute_saxs(q,flags,params):
    """Compute a SAXS intensity spectrum given some parameters.

    TODO: Document the equation.

    Parameters
    ----------
    q : array
        Array of q values at which saxs intensity should be computed.
    flags : dict
        Flags for scatterer populations. 
        Supported flags:
        - 'bad_data'
        - 'precursor_scattering' 
        - 'form_factor_scattering'
        - 'diffraction_peaks'
    params : dict
        Scattering equation parameters.
        Supported parameter keys:
        - 'I_at_0': intensity at q=0, by fitting the low-q region 
            to a polynomial with dI/dq(q=0)=0
        - 'I0_floor': magnitude of constant (flat in q) floor term 
        - 'rg_precursor': radius of gyration of precursors 
        - 'G_precursor': Guinier prefactor for precursor scattering 
        - 'I0_sphere': spherical form factor scattering intensity scaling factor 
        - 'r0_sphere': mean sphere size (Angstrom) 
        - 'sigma_sphere': fractional standard deviation of sphere size 

    Returns
    ------- 
    I : array
        Array of scattering intensities for each of the input q values
    """
    b_flag = flags['bad_data']
    s_flag = flags['diffraction_peaks']
    I = np.zeros(len(q))
    if not b_flag and not s_flag:
        pre_flag = flags['precursor_scattering']
        f_flag = flags['form_factor_scattering']
        I0_floor = params['I0_floor'] 
        I = I0_floor*np.ones(len(q))
        if pre_flag:
            rg_pre = params['rg_precursor']
            G_pre = params['G_precursor']
            # Precursor as a monodisperse spherical form factor:
            #I_pre = spherical_normal_saxs(q,r0_pre,0)
            # TODO: Finish implementing Guinier-Porod here, or similar 
            # TODO: implement G_pre as a (fixed?) parameter
            # Precursor as a Guinier-Porod equation for a small sphere:
            I_pre = guinier_porod(q,rg_pre,4,G_pre)
            I += I_pre
        if f_flag:
            I0_sph = params['I0_sphere']
            r0_sph = params['r0_sphere']
            sigma_sph = params['sigma_sphere']
            I_sph = spherical_normal_saxs(q,r0_sph,sigma_sph)
            I += I0_sph*I_sph
    return I

def spherical_normal_saxs(q,r0,sigma):
    """Compute SAXS intensity of a normally-distributed sphere population.

    Originally contributed by Amanda Fournier.

    The returned intensity is normalized 
    such that I(q=0) is equal to 1.

    Parameters
    ----------
    q : array
        array of scattering vector magnitudes
    r0 : float
        mean radius of the sphere population
    sigma : float
        fractional standard deviation of the sphere population radii

    Returns
    -------
    I : array
        Array of scattering intensities for each of the input q values
    """
    q_zero = (q == 0)
    q_nz = np.invert(q_zero) 
    I = np.zeros(q.shape)
    if sigma < 1E-9:
        x = q*r0
        V_r0 = float(4)/3*np.pi*r0**3
        I[q_nz] = V_r0**2 * (3.*(np.sin(x[q_nz])-x[q_nz]*np.cos(x[q_nz]))*x[q_nz]**-3)**2
        I_zero = V_r0**2 
    else:
        sigma_r = sigma*r0
        dr = sigma_r*0.02
        rmin = np.max([r0-5*sigma_r,dr])
        rmax = r0+5*sigma_r
        I_zero = 0
        for ri in np.arange(rmin,rmax,dr):
            xi = q*ri
            V_ri = float(4)/3*np.pi*ri**3
            # The normal-distributed density of particles with radius r_i:
            rhoi = 1./(np.sqrt(2*np.pi)*sigma_r)*np.exp(-1*(r0-ri)**2/(2*sigma_r**2))
            I_zero += V_ri**2 * rhoi*dr
            I[q_nz] += V_ri**2 * rhoi*dr*(3.*(np.sin(xi[q_nz])-xi[q_nz]*np.cos(xi[q_nz]))*xi[q_nz]**-3)**2
    if any(q_zero):
        I[q_zero] = I_zero
    I = I/I_zero 
    return I

def guinier_porod(q,r_g,porod_exponent,guinier_factor):
    """Compute the Guinier-Porod small-angle scattering intensity.
    
    Computes the Guinier-Porod scattering intensity,
    given the Guinier prefactor of the solvent/scatterer system,
    the Porod exponent of the scatterer geometry,
    and the radius of gyration of the scatterer.

    Reference
    ---------
    B. Hammouda, J. Appl. Cryst. (2010). 43, 716-719.
    """
    # q-domain boundary q_splice:
    q_splice = 1./r_g * np.sqrt(3./2*porod_exponent**2)
    idx_guinier = (q <= q_splice)
    idx_porod = (q > q_splice)
    # porod prefactor D:
    porod_factor = guinier_factor*np.exp(-1./2*porod_exponent)\
                    * (3./2*porod_exponent)**(1./2*porod_exponent)\
                    * 1./(r_g**porod_exponent)
    I = np.zeros(q.shape)
    # Guinier equation:
    if any(idx_guinier):
        I[idx_guinier] = guinier_factor * np.exp(-1./3*q[idx_guinier]**2*r_g**2)
    # Porod equation:
    if any(idx_porod):
        I[idx_porod] = porod_factor * 1./(q[idx_porod]**porod_exponent)
    return I


def profile_spectrum(q_I):
    """Numerical profiling of a SAXS spectrum.

    Profile a saxs spectrum (n-by-2 array q_I) 
    by taking several fast numerical metrics 
    from the measured data.
    The metrics should be consistent for spectra
    with different intensity scaling 
    or different q domains.   

    This method should execute gracefully
    for any n-by-2 input array,
    such that it can be used to profile any type of spectrum. 
    TODO: document the returned metrics here.

    Parameters
    ----------
    q_I : array
        n-by-2 array of scattering vector q and scattered intensity I
    
    Returns
    -------
    params : dict
        dictionary of scattering equation parameters,
        for input to compute_saxs() 
    """ 
    q = q_I[:,0]
    I = q_I[:,1]
    # I metrics
    idxmax = np.argmax(I)
    idxmin = np.argmin(I)
    I_min = I[idxmin]
    I_max = I[idxmax] 
    q_Imax = q[idxmax]
    I_range = I_max - I_min
    #I_sum = np.sum(I)
    I_mean = np.mean(I)
    Imax_over_Imean = I_max/I_mean
    # log(I) metrics
    nz = I>0
    q_nz = q[nz]
    I_nz = I[nz]
    logI_nz = np.log(I_nz)
    logI_max = np.max(logI_nz)
    logI_min = np.min(logI_nz)
    logI_range = logI_max - logI_min
    logI_std = np.std(logI_nz)
    logI_max_over_std = logI_max / logI_std
    # I_max peak shape analysis
    idx_around_max = ((q > 0.9*q_Imax) & (q < 1.1*q_Imax))
    Imean_around_max = np.mean(I[idx_around_max])
    Imax_sharpness = I_max / Imean_around_max

    ### fluctuation analysis
    # array of the difference between neighboring points:
    nn_diff = logI_nz[1:]-logI_nz[:-1]
    # keep indices where the sign of this difference changes.
    # also keep first index
    nn_diff_prod = nn_diff[1:]*nn_diff[:-1]
    idx_keep = np.hstack((np.array([True]),nn_diff_prod<0))
    fluc = np.sum(np.abs(nn_diff[idx_keep]))
    logI_fluctuation = fluc/logI_range

    # TODO: add some pearson metrics.
    # TODO: add qmin, qmax as metrics

    params = OrderedDict()
    params['q_Imax'] = q_Imax
    params['Imax_over_Imean'] = Imax_over_Imean
    params['Imax_sharpness'] = Imax_sharpness
    params['logI_fluctuation'] = logI_fluctuation
    params['logI_max_over_std'] = logI_max_over_std
    return params 

def parameterize_spectrum(q_I,flags):
    """Determine scattering equation parameters for a given spectrum.

    Parameters
    ----------
    q_I : array
        n-by-2 array of q (scattering vector in 1/A) and I (intensity)
    flags : dict
        dictionary of population flags, like SaxsClassifier.classify() output

    Returns
    -------
    params : dict
        dict of scattering equation parameters
        corresponding to the flagged populations,
        similar to the input params for compute_saxs().
    """
    #print('parameterizing for: {}'.format(flags))
    d = OrderedDict()
    #if flags['bad_data']:
    #    return d 
    q = q_I[:,0]
    I = q_I[:,1]
    n_q = len(q)
    # Get a number for I(q=0)
    # TODO: add units to this package
    # Disregard lowest-q values if they are far from the mean, 
    # as these points are likely dominated by experimental error,
    # such as interference from beam stops, etc.
    idx_lowq = (q<0.06)
    Imean_lowq = np.mean(I[idx_lowq])
    Istd_lowq = np.std(I[idx_lowq])
    idx_good = ((I[idx_lowq] < Imean_lowq+Istd_lowq) & (I[idx_lowq] > Imean_lowq-Istd_lowq))
    I_at_0 = fit_I0(q[idx_lowq][idx_good],I[idx_lowq][idx_good],4)
    # safety check in case of low-q errors (e.g. beam stop interference, beam drift)...
    if not I_at_0 > 0 or I_at_0 < 0.5*Imean_lowq: 
        #print('TRYING TO GET BETTER I0')
        idx_fit = ((q>0.06)&(q<0.1))
        I_at_0 = fit_I0(q[idx_fit],I[idx_fit],2)
    #elif flags['precursor_scattering']:
    #    I_at_0 = fit_I0(q,I,3)
    d['I_at_0'] = I_at_0

    #TODO: add parameters for diffraction peaks
    #if flags['diffraction_peaks']:
    #    .... 

    if flags['form_factor_scattering']:
        # TODO: insert cases for non-spherical form factors
        r0_sphere, sigma_sphere = spherical_normal_heuristics(
            np.array(zip(q,I)),I_at_0)
        #I_nz = np.invert((I<=0))
        #I_error = lambda x: np.sum( 
        #    (x*spherical_normal_saxs(q,r0_sphere,sigma_sphere) - I)**2 )
        #res = scipimin(I_error,[1E-6],bounds=[(0,None)])
        #I0_sphere = res.x[0]
        #I_sphere = I0_sphere*spherical_normal_saxs(q,r0_sphere,sigma_sphere)
        d['r0_sphere'] = r0_sphere
        d['sigma_sphere'] = sigma_sphere

    if flags['precursor_scattering']: 
        rg_pre, G_pre = precursor_heuristics(
            np.array(zip(q,I)))
        #idx_highq = (q > q[int(n_q*3./4)]) 
        #I_error = lambda x: np.sum( 
        #    (x*guinier_porod(q,rg_pre,4,G_pre)[idx_highq] - I[idx_highq])**2 )
        #res = scipimin(I_error,[1E-6],bounds=[(0,None)])
        #I0_pre = res.x[0]
        #I_pre = I0_pre*guinier_porod(q,rg_pre,4,G_pre)
        I_pre = guinier_porod(q,rg_pre,4,G_pre)
        d['rg_precursor'] = rg_pre 
        d['G_precursor'] = G_pre 

    #from matplotlib import pyplot as plt
    #plt.figure(1)
    #plt.plot(q,I)
    #plt.plot(q,I_sphere,'g')
    #plt.show()

    #I_floor = np.ones(n_q)
    #I_pre = np.zeros(n_q)
    #I_form = np.zeros(n_q)
    #if flags['precursor_scattering']: 
    #    I_pre = guinier_porod(q,rg_pre,4,G_pre)
    #if flags['form_factor_scattering']:
    #    I_form = spherical_normal_saxs(q,r0_sphere,sigma_sphere)
    #if flags['diffraction_peaks']:
    #   I_peaks = .....

    # Run a quick least-squares to get initial guesses for intensity prefactors
    #if flags['precursor_scattering'] and flags['form_factor_scattering']:
    #    x_init = [0.0,I_at_0]
    #    x_bounds = [(0.0,None),(1E-9,None)]
    #    Ifunc = lambda x: x[0]*I_floor + I_pre + x[1]*I_form 
    #elif flags['precursor_scattering']:
    #    x_init = [0.0]
    #    x_bounds = [(0.0,None)]
    #    Ifunc = lambda x: x[0]*I_floor + I_pre 
    f_pre = flags['precursor_scattering']
    f_form = flags['form_factor_scattering']
    f_pks = flags['diffraction_peaks']
    n_q = len(q)
    I_nz = np.invert((I<=0))
    if f_pre and f_form: 
        Ifunc = lambda x: x[0]*np.ones(n_q) \
            + guinier_porod(q,rg_pre,4,x[1]) \
            + x[2]*spherical_normal_saxs(q,r0_sphere,sigma_sphere) 
        I_error = lambda x: np.sum( (np.log(Ifunc(x)[I_nz]) - np.log(I[I_nz]))**2 )
        res = scipimin(I_error,[0.0,G_pre,I_at_0],
            bounds=[(0.0,None),(1E-6,None),(1E-6,None)],
            constraints=[{'type':'eq','fun':lambda x:np.sum(x)-I_at_0}])
        d['I0_floor'] = res.x[0]
        d['G_precursor'] = res.x[1]
        d['I0_sphere'] = res.x[2]
    elif f_pre:
        d['I0_floor'] = 0
        #Ifunc = lambda x: x[0]*np.ones(n_q) \
        #    + guinier_porod(q,rg_pre,4,x[1]) 
        #I_error = lambda x: np.sum( (np.log(Ifunc(x)[I_nz]) - np.log(I[I_nz]))**2 )
        #res = scipimin(I_error,[0.0,G_pre],bounds=[(0.0,None),(0.0,None)])
        #d['I0_floor'] = res.x[0]
        #d['G_precursor'] = res.x[1]
    elif f_form:
        I0_floor = np.mean(I[int(n_q*0.9):])
        d['I0_floor'] = I0_floor 
        d['I0_sphere'] = I_at_0 - I0_floor
        #Ifunc = lambda x: x[0]*np.ones(n_q) \
        #    + x[1]*spherical_normal_saxs(q,r0_sphere,sigma_sphere) 
        #I_error = lambda x: np.sum( (np.log(Ifunc(x)[I_nz]) - np.log(I[I_nz]))**2 )
        #res = scipimin(I_error,[0.0,I_at_0],bounds=[(0.0,None),(0.0,None)])
        #d['I0_floor'] = res.x[0]
        #d['I0_sphere'] = res.x[1]

        #if flags['precursor_scattering'] and flags['form_factor_scattering']:
        #    #I0_pre = res.x[1] 
        #    I0_sphere = res.x[2]
        #    #d['I0_precursor'] = I0_pre 
        #    d['I0_sphere'] = I0_sphere 
        #elif flags['precursor_scattering']:
        #    I0_pre = res.x[1]
        #    d['I0_precursor'] = I0_pre 
        #elif flags['form_factor_scattering']:
        #    I0_form = res.x[1]
        #    d['I0_sphere'] = I0_form 
        #d['I0_floor'] = I0_floor 
    I_guess = compute_saxs(q,flags,d)
    q_I_guess = np.array([q,I_guess]).T
    nz = ((I>0)&(I_guess>0))
    logI_nz = np.log(I[nz])
    logIguess_nz = np.log(I_guess[nz])
    Imean = np.mean(logI_nz)
    Istd = np.std(logI_nz)
    logI_nz_s = (logI_nz - Imean) / Istd
    logIguess_nz_s = (logIguess_nz - Imean) / Istd
    #d['R2log_guess'] = compute_Rsquared(logI_nz,logIguess_nz)
    #d['chi2log_guess'] = compute_chi2(logI_nz_s,logIguess_nz_s)
    return d

def fit_spectrum(q_I,flags,params,fixed_params,objective='chi2log'):
    """Fit a SAXS spectrum, given population flags and parameter guesses.

    Parameters
    ----------
    q_I : array
        n-by-2 array of scattering vector q (1/Angstrom) and intensity.
    flags : dict
        dictionary of flags indicating various scatterer populations.
    params : dict
        scattering equation parameters to use as intial condition for fitting.
    fixed_params : list
        list of keys (strings) indicating which entries in `params` 
        should be held constant during the optimization. 
    objective : str
        choice of objective function for the fitting. supported objectives:
        - 'chi2log': sum of difference of logarithm, squared, across entire q range. 
        - 'chi2log_fixI0': like chi2log, but with I(q=0) constrained. 

    Returns
    -------
    params_opt : dict
        Dict of scattering equation parameters 
        optimized to fit `q_I` under `objective`.
    """

    f_bd = flags['bad_data']
    f_pre = flags['precursor_scattering']
    f_form = flags['form_factor_scattering']
    f_pks = flags['diffraction_peaks']
    q = q_I[:,0]
    n_q = len(q)
    I = q_I[:,1]

    # trim non-flagged populations out of fit_params:
    # fit as few of the requested fit_params as possible
    #if not f_pre:
    #    if 'rg_precursor' in fit_params:
    #        fit_params.pop(fit_params.index('rg_precursor')) 
    #    if 'G_precursor' in fit_params:
    #        fit_params.pop(fit_params.index('G_precursor')) 
    #if not f_form:
    #    if 'I0_sphere' in fit_params:
    #        fit_params.pop(fit_params.index('I0_sphere')) 
    #    if 'r0_sphere' in fit_params:
    #        fit_params.pop(fit_params.index('r0_sphere')) 
    #    if 'sigma_sphere' in fit_params:
    #        fit_params.pop(fit_params.index('sigma_sphere')) 

    # index the params into an array 
    x_init = np.zeros(len(params))
    x_bounds = [] 
    I_keys = ['I0_floor','G_precursor','I0_sphere']
    p_idx = {} 
    I_idx = {} 
    for i,k in zip(range(len(params)),params.keys()):
        p_idx[k] = i 
        x_init[i] = params[k]
        if k in I_keys:
            I_idx[k] = i
        if k in ['rg_precursor','r0_sphere']:
            x_bounds.append((1E-3,None))
        elif k in ['G_precursor','I0_sphere','I0_floor']:
            x_bounds.append((0.0,None))
        elif k in ['sigma_sphere']:
            x_bounds.append((0.0,0.45))
        else:
            x_bounds.append((None,None))
    
    c = []
    if objective in ['chi2log_fixI0']:
        if len(I_idx) > 0:
            # Set up a constraint to keep I(q=0) fixed
            I0_init = np.sum([x_init[i] for i in I_idx])
            cfun = lambda x: np.sum([x[i] for i in I_idx] - I0_init)
            c.append({'type':'eq','fun':cfun})

    for fixk in fixed_params:
        cfun = lambda x: x[p_idx[fixk]] - params[fixk]
        c.append({'type':'eq','fun':cfun})

    p_opt = copy.deepcopy(params) 
    # Only proceed if there is work to do.
    if not f_pks and not f_bd:
        saxs_fun = lambda q,x,p: compute_saxs_with_substitutions(q,flags,p,params.keys(),x)
        if objective in ['chi2log','chi2log_fixI0']:
            #idx_fit = ((I>0)&(q>0.1))
            if f_pre and not f_form and not f_pks:
                # if we are fitting only precursors,
                # intensities should be relatively low,
                # so low-q region can be problematic. 
                # TODO: consider whether this special case is justified,
                # or come up with a better way to filter high-error low-q points
                idx_fit = ((I>0)&(np.arange(len(q))>n_q*1./4))
            else:
                idx_fit = (I>0)
            q_fit = q[idx_fit]
            I_fit = I[idx_fit]
            Imean_fit = np.mean(I_fit)
            Istd_fit = np.std(I_fit)
            Is_fit = (I_fit-Imean_fit)/Istd_fit
            fit_obj = lambda x: compute_chi2(
            (np.log(saxs_fun(q_fit,x,params))-Imean_fit)/Istd_fit , Is_fit)
        p_opt['objective_before'] = fit_obj(x_init)
        res = scipimin(fit_obj,x_init,
            method='SLSQP',bounds=x_bounds,
            options={'ftol':1E-3},constraints=c)
        p_opt['objective_after'] = fit_obj(res.x)
        for k,xk in zip(params.keys(),res.x):
            p_opt[k] = xk
        p_opt['fixed_params'] = fixed_params
        p_opt['objective'] = objective 

    #print('precursors: {}, spheres: {}, peaks: {}'.format(f_pre,f_form,f_pks))
    #print('params: {}'.format(d_opt))

    #I_guess = compute_saxs(q,flags,params) 
    #I_opt = compute_saxs(q,flags,d_opt) 
    
    #from matplotlib import pyplot as plt
    #plt.figure(2)
    #plt.plot(q,I)
    #plt.plot(q,I_guess,'r')
    #plt.plot(q,I_opt,'g')
    #plt.figure(12)
    #plt.semilogy(q,I)
    #plt.semilogy(q,I_guess,'r')
    #plt.semilogy(q,I_opt,'g')
    #plt.show()

    return p_opt    

        #if d_opt['objective_after'] > d_opt['objective_before']:
        #    print('WARNING: optimization has increased the objective function. why?')
        #    print('x_init: {}'.format(x_init))
        #    print('obj_init: {}'.format(d_opt['objective_before']))
        #    print('x_opt: {}'.format(x_opt))
        #    print('obj_opt: {}'.format(d_opt['objective_after']))
        #    x_opt = x_init


    #Supported objective functions: 
    #(1) 'chi2': sum of difference squared across entire q range. 
    #(3) 'chi2norm': sum of difference divided by measured value, squared, aross entire q range. 
    #(4) 'low_q_chi2': sum of difference squared in only the lowest half of measured q range. 
    #(5) 'low_q_chi2log': sum of difference of logarithm, squared, in lowest half of measured q range. 
    #(6) 'pearson': pearson correlation between measured and modeled spectra. 
    #(7) 'pearson_log': pearson correlation between logarithms of measured and modeled spectra.
    #(8) 'low_q_pearson': pearson correlation between measured and modeled spectra. 
    #(9) 'low_q_pearson_log': pearson correlation between logarithms of measured and modeled spectra. 

    #Supported constraints: 
    #(1) 'fix_I0': keeps I(q=0) fixed to the value specified in the input params.

    #TODO: document the objective functions, etc.
    #    I_nz = (I>0)
    #    n_q = len(q)
    #    idx_lowq = (q<0.4)
    #    if objfun == 'chi2':
    #        fit_obj = lambda x: compute_chi2(
    #        saxs_fun(q,x,params),I)
    #    elif objfun == 'chi2log':
    #        fit_obj = lambda x: compute_chi2(
    #        np.log(saxs_fun(q[I_nz],x,params)),
    #        np.log(I[I_nz]))
    #    elif objfun == 'chi2norm':
    #        fit_obj = lambda x: compute_chi2(
    #        saxs_fun(q[I_nz],x,params),
    #        I[I_nz] , weights=float(1)/I[I_nz])
    #    elif objfun == 'low_q_chi2':
    #        fit_obj = lambda x: compute_chi2(
    #        saxs_fun(q[idx_lowq],x,params),I[idx_lowq] )
    #    elif objfun == 'pearson':
    #        fit_obj = lambda x: -1*compute_pearson(
    #        saxs_fun(q,x,params),I)
    #    elif objfun == 'low_q_pearson':
    #        fit_obj = lambda x: -1*compute_pearson( 
    #        saxs_fun(q[idx_lowq],x,params),
    #        I[idx_lowq] )
    #    elif objfun == 'low_q_chi2log':
    #        fit_obj = lambda x: compute_chi2( 
    #        np.log(saxs_fun(q[I_nz][idx_lowq[I_nz]],x,params)), 
    #        np.log(I[I_nz][idx_lowq[I_nz]]) ) 
    #    elif objfun == 'pearson_log':
    #        fit_obj = lambda x: -1*compute_pearson( 
    #        np.log(saxs_fun(q[I_nz],x,params)),np.log(I[I_nz])) 
    #    elif objfun == 'low_q_pearson_log':
    #        fit_obj = lambda x: -1*compute_pearson( 
    #        np.log(saxs_fun(q[I_nz][idx_lowq[I_nz]],x,params)), 
    #        np.log(I[I_nz][idx_lowq[I_nz]]) ) 
    #    else:
    #        msg = 'objective function {} not supported'.format(objfun)
    #        raise ValueError(msg)
    #    #d_opt['R2log_fit'] = compute_Rsquared(np.log(I[I_nz]),np.log(I_opt[I_nz]))
    #    #d_opt['chi2log_fit'] = compute_chi2(logI_nz_s,logIopt_nz_s)
        #I_keys = []
        #if 'I0_floor' in fit_params:
        #    I_keys.append('I0_floor')
        #if 'G_precursor' in fit_params:
        #    I_keys.append('G_precursor')
        #if 'I0_sphere' in fit_params:
        #    I_keys.append('I0_sphere')
        ##if 'I0_structure' in fit_params:
        ##    I_keys.append('I0_structure')
        #if len(I_keys) == 0:
        #    # constraint inherently satisfied: do nothing.
        #    pass
        #elif len(I_keys) == 1:
        #    # satisfy constraint by removing this key from fit_params. 
        #    fit_params.pop(fit_params.index(I_keys[0])) 
        #else:
        #    # find the indices of the relevant fit_params and form a constraint function.
        #    iargs = []
        #    Icons = 0
        #    for ik,k in zip(range(len(fit_params)),fit_params):
        #        if k in I_keys:
        #            iargs.append(ik)
        #            Icons += params[k]
        #    # Constrain I(q=0) to be fixed during fit
        #    #constraints=[{'type':'eq','fun':lambda x:np.sum(x)-I_at_0}]) 
        #    cfun = lambda x: sum([x[i] for i in iargs]) - Icons 
        #    c_fixI0 = {'type':'eq','fun':cfun}
        #    x_constraints.append(c_fixI0)

def precursor_heuristics(q_I):
    """Guess radius of gyration and Guinier prefactor of scatterers.

    Parameters
    ----------
    q_I : array
        n-by-2 array of q (scattering vector magnitude) 
        and I (intensity at q)

    Returns
    -------
    rg_pre : float
        estimated radius of gyration 
    G_pre : float
        estimated Guinier factor
    """
    n_q = len(q_I[:,0])
    ## use the higher-q regions 
    highq_I = q_I[int(n_q*1./2):,:] 
    #highq_I = q_I[int(n_q*3./4):,:] 
    fit_obj = lambda x: fit_guinier_porod(highq_I,x[0],4,x[1])
    idx_nz = highq_I[:,1]>0
    #fit_obj = lambda x: -1*compute_pearson(
    #    np.log(guinier_porod(highq_I[idx_nz,0],x[0],4,x[1])),
    #    np.log( highq_I[idx_nz,1]) )
    ##res = scipimin(fit_ojb,[0.1],bounds=[(0,0.3)]) 
    res = scipimin(fit_obj,[1,1],bounds=[(1E-3,10),(1E-3,None)])
    rg_opt, G_opt = res.x
    I_pre = guinier_porod(q_I[:,0],rg_opt,4,G_opt)
    return rg_opt, G_opt

def fit_guinier_porod(q_I,rg,porod_exp,G):
    #I_mean = np.mean(q_I[:,1])
    #I_std = np.std(q_I[:,1])
    #I_max = np.max(q_I[:,1])
    #I_s = q_I[:,1] / I_max
    Igp = guinier_porod(q_I[:,0],rg,porod_exp,G)
    #Igp_s = (Igp-np.mean(Igp)) / np.std(Igp)
    return np.sum( (q_I[:,1]-Igp)**2 )

def spherical_normal_heuristics(q_I,I_at_0):
    """Guess mean and std of radii for spherical scatterers.

    This algorithm was developed and 
    originally contributed by Amanda Fournier.    

    Performs some heuristic measurements on the input spectrum,
    in order to make educated guesses 
    for the parameters of a size distribution
    (mean and standard deviation of radius)
    for a population of spherical scatterers.

    TODO: Document algorithm here.
    """
    m = saxs_Iq4_metrics(q_I)

    width_metric = m['pI_qwidth']/m['q_at_Iqqqq_min1']
    intensity_metric = m['I_at_Iqqqq_min1']/I_at_0
    #######
    #
    # The function spherical_normal_heuristics_setup()
    # (in this same module) should be used to regenerate these polynomials
    # if any part of saxs_Iq4_metrics() is changed.
    # polynomial coefs for qr0 focus: 
    p_f = [16.86239254,8.85709143,-11.10439599,-0.26735688,4.49884714]
    # polynomial coefs for width metric: 
    p_w = [12.42148677,-16.85723287,7.43401497,-0.38234993,0.06203096]
    # polynomial coefs for intensity metric: 
    p_I = [1.19822603,-1.20386273,2.88652860e-01,1.78677430e-02,-2.67888841e-04]
    #
    #######
    # Find the sigma_r/r0 value that gets the extracted metrics
    # as close as possible to p_I and p_w.
    width_error = lambda x: (np.polyval(p_w,x)-width_metric)**2
    intensity_error = lambda x: (np.polyval(p_I,x)-intensity_metric)**2
    # TODO: make the objective function weight all errors equally
    heuristics_error = lambda x: width_error(x) + intensity_error(x)
    res = scipimin(heuristics_error,[0.1],bounds=[(0,0.45)]) 
    sigma_over_r = res.x[0]
    qr0_focus = np.polyval(p_f,sigma_over_r)
    # qr0_focus = x1  ==>  r0 = x1 / q1
    r0 = qr0_focus/m['q_at_Iqqqq_min1']
    #print('r0: {}, sigma/r: {}'.format(r0,sigma_over_r))
    return r0,sigma_over_r

def saxs_Iq4_metrics(q_I):
    """
    From an input spectrum q and I(q),
    compute several properties of the I(q)*q^4 curve.
    This was designed for spectra that are 
    dominated by a dilute spherical form factor term.
    The metrics extracted by this Operation
    were originally intended as an intermediate step
    for estimating size distribution parameters 
    for a population of dilute spherical scatterers.

    Returns a dict of metrics.
    Dict keys and meanings:
    q_at_Iqqqq_min1: q value at first minimum of I*q^4
    I_at_Iqqqq_min1: I value at first minimum of I*q^4
    Iqqqq_min1: I*q^4 value at first minimum of I*q^4
    pIqqqq_qwidth: Focal q-width of polynomial fit to I*q^4 near first minimum of I*q^4 
    pIqqqq_Iqqqqfocus: Focal point of polynomial fit to I*q^4 near first minimum of I*q^4
    pI_qvertex: q value of vertex of polynomial fit to I(q) near first minimum of I*q^4  
    pI_Ivertex: I(q) at vertex of polynomial fit to I(q) near first minimum of I*q^4
    pI_qwidth: Focal q-width of polynomial fit to I(q) near first minimum of I*q^4
    pI_Iforcus: Focal point of polynomial fit to I(q) near first minimum of I*q^4

    TODO: document the algorithm here.
    """
    q = q_I[:,0]
    I = q_I[:,1]
    d = {}
    #if not dI:
    #    # uniform weights
    #    wt = np.ones(q.shape)   
    #else:
    #    # inverse error weights, 1/dI, 
    #    # appropriate if dI represents
    #    # Gaussian uncertainty with sigma=dI
    #    wt = 1./dI
    #######
    # Heuristics step 1: Find the first local max
    # and subsequent local minimum of I*q**4 
    Iqqqq = I*q**4
    # w is the number of adjacent points to consider 
    # when examining the I*q^4 curve for local extrema.
    # A greater value of w filters out smaller extrema.
    w = 10
    idxmax1, idxmin1 = 0,0
    stop_idx = len(q)-w-1
    test_range = iter(range(w,stop_idx))
    idx = test_range.next() 
    while any([idxmax1==0,idxmin1==0]) and idx < stop_idx-1:
        if np.argmax(Iqqqq[idx-w:idx+w+1]) == w and idxmax1 == 0:
            idxmax1 = idx
        if np.argmin(Iqqqq[idx-w:idx+w+1]) == w and idxmin1 == 0 and not idxmax1 == 0:
            idxmin1 = idx
        idx = test_range.next()
    if idxmin1 == 0 or idxmax1 == 0:
        ex_msg = str('unable to find first maximum and minimum of I*q^4 '
        + 'by scanning for local extrema with a window width of {} points'.format(w))
        d['message'] = ex_msg 
        raise RuntimeError(ex_msg)
    #######
    # Heuristics 2: Characterize I*q**4 around idxmin1, 
    # by locally fitting a standardized polynomial.


    idx_around_min1 = (q>0.9*q[idxmin1]) & (q<1.1*q[idxmin1])
    # keep only the lower-q side, to encourage upward curvature
    #idx_around_min1 = (q>0.8*q[idxmin1]) & (q<q[idxmin1])


    q_min1_mean = np.mean(q[idx_around_min1])
    q_min1_std = np.std(q[idx_around_min1])
    q_min1_s = (q[idx_around_min1]-q_min1_mean)/q_min1_std
    Iqqqq_min1_mean = np.mean(Iqqqq[idx_around_min1])
    Iqqqq_min1_std = np.std(Iqqqq[idx_around_min1])
    Iqqqq_min1_s = (Iqqqq[idx_around_min1]-Iqqqq_min1_mean)/Iqqqq_min1_std
    #Iqqqq_min1_quad = lambda x: np.sum((x[0]*q_min1_s**2 + x[1]*q_min1_s + x[2] - Iqqqq_min1_s)**2)
    #res = scipimin(Iqqqq_min1_quad,[1E-3,0,0],bounds=[(0,None),(None,None),(None,None)])
    #p_min1 = res.x
    p_min1 = np.polyfit(q_min1_s,Iqqqq_min1_s,2,None,False,np.ones(len(q_min1_s)),False)
    # polynomial vertex horizontal coord is -b/2a
    qs_at_min1 = -1*p_min1[1]/(2*p_min1[0])
    d['q_at_Iqqqq_min1'] = qs_at_min1*q_min1_std+q_min1_mean
    # polynomial vertex vertical coord is poly(-b/2a)
    Iqqqqs_at_min1 = np.polyval(p_min1,qs_at_min1)
    d['Iqqqq_min1'] = Iqqqqs_at_min1*Iqqqq_min1_std+Iqqqq_min1_mean
    d['I_at_Iqqqq_min1'] = d['Iqqqq_min1']*float(1)/(d['q_at_Iqqqq_min1']**4)
    # The focal width of the parabola is 1/a 
    p_min1_fwidth = abs(1./p_min1[0])
    d['pIqqqq_qwidth'] = p_min1_fwidth*q_min1_std
    # The focal point is at -b/2a,poly(-b/2a)+1/(4a)
    p_min1_fpoint = Iqqqqs_at_min1+float(1)/(4*p_min1[0])
    d['pIqqqq_Iqqqqfocus'] = p_min1_fpoint*Iqqqq_min1_std+Iqqqq_min1_mean
    #######
    # Heuristics 2b: Characterize I(q) near min1 of I*q^4.
    I_min1_mean = np.mean(I[idx_around_min1])
    I_min1_std = np.std(I[idx_around_min1])
    I_min1_s = (I[idx_around_min1]-I_min1_mean)/I_min1_std
    #I_min1_error = lambda x: np.sum((x[0]*q_min1_s**2 + x[1]*q_min1_s + x[2] - I_min1_s)**2)
    #res = scipimin(I_min1_error,[0,0,0],bounds=[(0,None),(None,None),(None,None)])
    #pI_min1 = res.x
    pI_min1 = np.polyfit(q_min1_s,I_min1_s,2,None,False,np.ones(len(q_min1_s)),False)
    # polynomial vertex horizontal coord is -b/2a
    qs_vertex = -1*pI_min1[1]/(2*pI_min1[0])
    d['pI_qvertex'] = qs_vertex*q_min1_std+q_min1_mean
    # polynomial vertex vertical coord is poly(-b/2a)
    Is_vertex = np.polyval(pI_min1,qs_vertex)
    d['pI_Ivertex'] = Is_vertex*I_min1_std+I_min1_mean
    # The focal width of the parabola is 1/a 
    pI_fwidth = abs(1./pI_min1[0])
    d['pI_qwidth'] = pI_fwidth*q_min1_std
    # The focal point is at -b/2a,poly(-b/2a)+1/(4a)
    pI_fpoint = Is_vertex+float(1)/(4*pI_min1[0])
    d['pI_Ifocus'] = pI_fpoint*I_min1_std+I_min1_mean
    #######
    return d

def compute_saxs_with_substitutions(q,flags,params,x_keys,x_vals):
    p_sub = copy.deepcopy(params)
    for k,v in zip(x_keys,x_vals):
        p_sub[k] = copy.copy(v)
    return compute_saxs(q,flags,p_sub)

def compute_chi2(y1,y2,weights=None):
    """
    Compute the sum of the difference squared between input arrays y1 and y2.
    """
    if weights is None:
        return np.sum( (y1 - y2)**2 )
    else:
        weights = weights / np.sum(weights)
        return np.sum( (y1 - y2)**2*weights )

def compute_Rsquared(y1,y2):
    """
    Compute the coefficient of determination between input arrays y1 and y2.
    """
    sum_var = np.sum( (y1-np.mean(y1))**2 )
    sum_res = np.sum( (y1-y2)**2 ) 
    return float(1)-float(sum_res)/sum_var

def compute_pearson(y1,y2):
    """
    Compute the Pearson correlation coefficient between input arrays y1 and y2.
    """
    y1mean = np.mean(y1)
    y2mean = np.mean(y2)
    y1std = np.std(y1)
    y2std = np.std(y2)
    return np.sum((y1-y1mean)*(y2-y2mean))/(np.sqrt(np.sum((y1-y1mean)**2))*np.sqrt(np.sum((y2-y2mean)**2)))

def fit_I0(q,I,order=4):
    """
    Find an estimate for I(q=0) by polynomial fitting.
    All of the input q, I(q) values are used in the fitting.
    """
    #TODO: add a sign constraint, at least
    I_mean = np.mean(I)
    I_std = np.std(I)
    q_mean = np.mean(q)
    q_std = np.std(q)
    I_s = (I-I_mean)/I_std
    q_s = (q-q_mean)/q_std
    p = fit_with_slope_constraint(q_s,I_s,-1*q_mean/q_std,0,order) 
    I_at_0 = np.polyval(p,-1*q_mean/q_std)*I_std+I_mean
    return I_at_0

def fit_with_slope_constraint(q,I,q_cons,dIdq_cons,order,weights=None):
    """
    Perform a polynomial fitting 
    of the low-q region of the spectrum
    with dI/dq(q=0) constrained to be zero.
    This is performed by forming a Lagrangian 
    from a quadratic cost function 
    and the Lagrange-multiplied constraint function.
    
    TODO: Explicitly document cost function, constraints, Lagrangian.

    Inputs q and I are not standardized in this function,
    so they should be standardized beforehand 
    if standardized fitting is desired.
    At the provided constraint point, q_cons, 
    the returned polynomial will have slope dIdq_cons.

    Because of the form of the Lagrangian,
    this constraint cannot be placed at exactly zero.
    This would result in indefinite matrix elements.
    """
    Ap = np.zeros( (order+1,order+1),dtype=float )
    b = np.zeros(order+1,dtype=float)
    # TODO: vectorize the construction of Ap
    for i in range(0,order):
        for j in range(0,order):
            Ap[i,j] = np.sum( q**j * q**i )
        Ap[i,order] = -1*i*q_cons**(i-1)
    for j in range(0,order):
        Ap[order,j] = j*q_cons**(j-1)
        b[j] = np.sum(I*q**j)
    b[order] = dIdq_cons
    p_fit = np.linalg.solve(Ap,b) 
    p_fit = p_fit[:-1]  # throw away Lagrange multiplier term 
    p_fit = p_fit[::-1] # reverse coefs to get np.polyfit format
    #from matplotlib import pyplot as plt
    #plt.figure(3)
    #plt.plot(q,I)
    #plt.plot(q,np.polyval(p_fit,q))
    #plt.plot(np.arange(q_cons,q[-1],q[-1]/100),np.polyval(p_fit,np.arange(q_cons,q[-1],q[-1]/100)))
    #plt.plot(q_cons,np.polyval(p_fit,q_cons),'ro')
    #plt.show()
    return p_fit
 
def spherical_normal_heuristics_setup():
    sigma_over_r = []
    width_metric = []
    intensity_metric = []
    qr0_focus = []
    # TODO: generate heuristics on a 1-d grid of sigma/r
    # instead of the 2-d grid being used here now.
    # Document proof that a 1-d grid is sufficient.
    #r0_vals = np.arange(10,41,10,dtype=float)              #Angstrom
    #for ir,r0 in zip(range(len(r0_vals)),r0_vals):
    r0 = 10
    q = np.arange(0.001/r0,float(20)/r0,0.001/r0)       #1/Angstrom
    # NOTE: algorithm works for sigma/r up to 0.45
    sigma_r_vals = np.arange(0*r0,0.46*r0,0.01*r0)      #Angstrom
    for isig,sigma_r in zip(range(len(sigma_r_vals)),sigma_r_vals):
        I = spherical_normal_saxs(q,r0,sigma_r/r0) 
        I_at_0 = spherical_normal_saxs(np.array([0]),r0,sigma_r/r0)[0] 
        print('getting I*q**4 metrics for sigma_r/r0 = {}'.format(sigma_r/r0))
        d = saxs_Iq4_metrics(np.array(zip(q,I)))
        sigma_over_r.append(float(sigma_r)/r0)
        qr0_focus.append(d['q_at_Iqqqq_min1']*r0)
        width_metric.append(d['pI_qwidth']/d['q_at_Iqqqq_min1'])
        intensity_metric.append(d['I_at_Iqqqq_min1']/I_at_0)
    # TODO: standardize before fitting, then revert after
    p_f = np.polyfit(sigma_over_r,qr0_focus,4,None,False,None,False)
    p_w = np.polyfit(sigma_over_r,width_metric,4,None,False,None,False)
    p_I = np.polyfit(sigma_over_r,intensity_metric,4,None,False,None,False)
    print('polynomial coefs for qr0 focus: {}'.format(p_f))
    print('polynomial coefs for width metric: {}'.format(p_w))
    print('polynomial coefs for intensity metric: {}'.format(p_I))
    plot = True
    if plot: 
        from matplotlib import pyplot as plt
        plt.figure(1)
        plt.scatter(sigma_over_r,width_metric)
        plt.plot(sigma_over_r,np.polyval(p_w,sigma_over_r))
        plt.figure(2)
        plt.scatter(sigma_over_r,intensity_metric)
        plt.plot(sigma_over_r,np.polyval(p_I,sigma_over_r))
        plt.figure(3)
        plt.scatter(sigma_over_r,qr0_focus)
        plt.plot(sigma_over_r,np.polyval(p_f,sigma_over_r))
        plt.figure(4)
        plt.scatter(width_metric,intensity_metric) 
        plt.show()






"""
Calibrate and reduce an image, given calibration parameters.

This module calls on pipeline.remesh.remesh 
to correct the (GI) images for curvature of 
the Ewald's sphere. 
"""

import numpy as np
from pyFAI.geometry import Geometry
# TODO: this op shouldn't depend on Xicam's pipeline module
#from pipeline import remesh

from ... import Operation as opmod 
from ...Operation import Operation

class Remesh(Operation):
    """
    Input image data (ndarray), pyFAI Geometory object, Angle of Incidence
    Return q_par, q_vrt, I(q_par, q_vrt) 
    """
    def __init__(self):
        input_names = ['image_data','pyFAI_geometory', 'alphai']
        output_names = ['qpar','qvrt','I']
        super(Remesh, self).__init__(input_names, output_names)
        self.input_doc['image_data'] = '2d array representing intensity for each pixel'
        self.input_doc['pyFAI_geometory'] = 'PyFAI Geometry python object'
        self.input_doc['alphai'] = 'Angle of Incidence'
        self.input_type['image_data'] = opmod.workflow_item
        self.input_type['pyFAI_geometory'] = opmod.workflow_item
        self.input_type['alphai'] = opmod.workflow_item
        self.output_doc['qpar'] = 'q-parallel'
        self.output_doc['qvrt'] = 'q-vertical' 
        self.output_doc['I'] = 'intensity as a function of qpar, qvrt'

    def run(self):
        img = self.inputs['image_data']
        geometry = self.inputs['pyFAI_geometory']
        alphai = self.inputs['alphai']
        #qpar, qvrt, intensity = remesh.remesh(img, geometry, alphai)
        qpar, qvrt, intensity = None, None, None
        # save results to self.outputs
        self.outputs['qpar'] = qpar
        self.outputs['qvrt'] = qvrt
        self.outputs['I'] = intensity



import os.path

import tifffile
import numpy as np

from ... import Operation as opmod 
from ...Operation import Operation

class LoadTif(Operation):
    """
    Takes a filesystem path that points to a .tif,
    outputs image data from the file. 
    """

    def __init__(self):
        input_names = ['file_path']
        output_names = ['image_data','dir_path','filename']
        super(LoadTif,self).__init__(input_names,output_names)
        self.input_doc['file_path'] = 'path to a .tif image'
        self.output_doc['image_data'] = '2D array representing pixel values'
        self.output_doc['filename'] = 'Filename for image, path and extension stripped'
        
    def run(self):
        p = self.inputs['file_path']
        if p is None:
            return
        dir_path = os.path.split(p)[0]
        file_nopath = os.path.split(p)[1]
        file_noext = os.path.splitext(file_nopath)[0]
        self.outputs['dir_path'] = dir_path 
        self.outputs['filename'] = file_noext 
        self.outputs['image_data'] = tifffile.imread(p)




import numpy as np

from ... import Operation as opmod
from ...Operation import Operation

class SpectrumClassifier(Operation):
    """Identifies scatterer populations from features of SAXS spectra."""

    def __init__(self):
        input_names = ['features', 'classifier']
        output_names = ['flags']
        super(SpectrumClassifier, self).__init__(input_names, output_names)
        self.input_doc['features'] = 'Dict of scalar features '\
            'as produced by PROCESSING.SAXS.SpectrumProfiler.'
        self.input_doc['classifier'] = 'A SaxsClassifier object, '\
            'as produced by IO.MODELS.SAXS.LoadSAXSClassifier.'
        self.output_doc['flags'] = 'Dict of boolean flags '\
            'indicating the presence of various scattering populations.'
        self.input_type['features'] = opmod.workflow_item
        self.input_type['classifier'] = opmod.workflow_item

    def run(self):
        x = self.inputs['features']
        c = self.inputs['classifier'] 
        # for py3, the dict.values() has to be cast as a list,
        # then the list has to be cast as a reshaped np.array
        # to be correctly understood by scikit-learn.
        self.outputs['flags'] = c.classify(np.array(list(x.values())).reshape(1,-1))




from ... import Operation as opmod 
from ...Operation import Operation
        
class CheckDataSet(Operation):
    """
    Take a Citrination client as input and use it to query a data set.
    Output some indication of whether or not the query was successful.
    """
    
    def __init__(self):
        input_names = ['client','dsid']
        output_names = ['ok_flag','status']
        super(CheckDataSet,self).__init__(input_names,output_names)
        self.input_doc['client'] = 'A reference to a running Citrination client.'
        self.input_doc['dsid'] = 'The data set to be queried.'
        self.output_doc['ok_flag'] = 'Indicator of whether or not the data set passes the test.'
        self.output_doc['status'] = 'Message describing the state of the data set.'
        self.input_type['client'] = opmod.plugin_item

    def run(self):
        c = self.inputs['client']
        dsid = self.inputs['dsid'] 
        if c is None or dsid is None:
            return
        f = True
        try:
            r = c.get_dataset_files(dsid)
            if 'name' in r.keys():
                s = 'client successfully queried data set {}: {}. Response: {}'.format(dsid,r['name'],r)
                f = True
            else:
                s = 'client queried data set {}, but was unable to find the data set name. Response: {}'.format(dsid,r)
                f = True
        except Exception as ex:
            s = 'client failed to query data set number {}. Error message: {}'.format(dsid,ex.message)
            f = False
        self.outputs['ok_flag'] = f
        self.outputs['status'] = s




import abc
from collections import OrderedDict

from ..operations import Operation as opmod

class PawsPlugin(object):
    __metaclass__ = abc.ABCMeta
    """
    Abstract class for imposing paws plugin structure via inheritance.
    """

    def __init__(self,input_names):
        super(PawsPlugin,self).__init__()
        self.inputs = OrderedDict()
        self.input_doc = {}
        # For each of the var names, assign to None 
        for name in input_names: 
            self.inputs[name] = None
            self.input_doc[name] = None

    def __getitem__(self,key):
        d = self.content()
        if key in d.keys():
            return d[key]
        else:
            raise KeyError('[{}] {}.__getitem__ only recognizes keys {}'
            .format(__name__,type(self).__name__,d.keys()))
    def __setitem__(self,key,data):
        d = self.content()
        if key in d.keys():
            d[key] = data
        else:
            raise KeyError('[{}] {}.__setitem__ only recognizes keys {}'
            .format(__name__,type(self).__name__,d.keys()))
    def keys(self):
        return self.content().keys() 

    def start(self):
        """
        PawsPlugin.start() should perform any setup required by the plugin,
        for instance setting up connections and reading files used by the plugin.
        The default implementation does nothing.
        """
        pass

    def stop(self):
        """
        PawsPlugin.stop() should provide a clean end for the plugin,
        for instance closing all connections and files used by the plugin.
        The default implementation does nothing, 
        assumes the plugin can be cleanly terminated by dereferencing. 
        """
        pass

    def description(self):
        """
        PawsPlugin.description() returns a string 
        documenting the functionality of the PawsPlugin.
        The default implementation returns no description. 
        """
        return "No description available"

    def content(self):
        """
        PawsPlugin.content() returns a dict
        containing the meaningful objects contained in the plugin.
        The default implementation returns an empty dict. 
        """
        return {}





import os
import numpy as np

from ... import Operation as opmod 
from ...Operation import Operation

class WriteArrayCSV(Operation):
    """Write a 2d array to a csv file"""

    def __init__(self):
        input_names = ['array','headers','dir_path','filename','filetag']
        output_names = ['file_path','filename']
        super(WriteArrayCSV, self).__init__(input_names, output_names)
        self.input_doc['array'] = 'any 2d array'
        self.input_doc['headers'] = 'list of string headers (optional)- one header for each column of array'
        self.input_doc['dir_path'] = 'the path to the destination directory'
        self.input_doc['filename'] = 'the name of the file to be saved- no extension is expected'
        self.input_doc['filetag'] = 'tag appended to filename- no extension is expected'
        self.output_doc['file_path'] = 'the path to the finished csv file: dir_path+filename+filetag+.csv'
        self.output_doc['file_path'] = 'the name of the output csv: filename+filetag'
        self.input_type['array'] = opmod.workflow_item
        self.inputs['filetag'] = ''

    def run(self):
        a = self.inputs['array']
        h = self.inputs['headers']
        p = self.inputs['dir_path']
        fnm = self.inputs['filename']
        tag = self.inputs['filetag']
        if p is None or a is None or fnm is None:
            return 
        csv_path = os.path.join(p,self.inputs['filename']+tag+'.csv')
        self.outputs['file_path'] = csv_path
        self.outputs['filename'] = fnm+tag 
        if h is not None:
            h_str = ''
            for i in range(len(h)-1):
                h_str += h[i] + ', '
            h_str = h_str+h[-1]
            np.savetxt(csv_path, a, delimiter=',', newline=os.linesep, header=h_str)
        else:
            np.savetxt(csv_path, a, delimiter=',', newline=os.linesep)




import os
from collections import OrderedDict

import tifffile

from ... import Operation as opmod 
from ...Operation import Operation

class ReadImageAndHeader_SSRL15(Operation):
    """
    Read an image and header generated by beamline 1-5 at SSRL.
    Returns ndarray image and dictionary header.
    """

    def __init__(self):
        input_names = ['file_path']
        output_names = ['image_data', 'image_header', 'dir_path', 'filename']
        super(ReadImageAndHeader_SSRL15, self).__init__(input_names, output_names)
        self.input_doc['file_path'] = 'path to a tif file '\
        'produced by beamline 1-5 at SSRL. '\
        'A .txt header file is expected '\
        'in the same directory as this .tif file.'
        self.output_doc['image_data'] = 'the image pixel data as an ndarray'
        self.output_doc['image_header'] = 'the header file as a python dictionary'
        self.output_doc['dir_path'] = 'the directory portion of the input file_path'
        self.output_doc['filename'] = 'filename with path and extension stripped'

    def run(self):
        tif_path = self.inputs['file_path']
        if tif_path is None:
            return 
        dirpath,filename = os.path.split(tif_path)
        filename_noext = os.path.splitext(filename)[0]
        path_noext = os.path.splitext(tif_path)[0]
        hdr_file_path = path_noext + '.txt'
        self.outputs['dir_path'] = dirpath 
        self.outputs['filename'] = filename_noext 
        self.outputs['image_data'] = tifffile.imread(tif_path)
        d = OrderedDict()
        for l in open(hdr_file_path,'r').readlines():
            try:
                if not l.strip() == '' and not l.strip()[0] == '#':
                    kvs = l.split(',')
                    # special case for the string headers on line 1
                    if not kvs[0].find('User') == -1:
                        u_str = kvs[0].split('User:')[1].strip()
                        t_str = kvs[1].split('time:')[1].strip()
                        d['User'] = u_str
                        d['time'] = t_str
                    # and filter out the redundant temperature line
                    elif not (len(kvs)==1 and kvs[0].strip()[-1]=='C'):
                        for kv in kvs:
                            kv_arr = kv.split('=')
                            d[kv_arr[0].strip()] = float(kv_arr[1].strip())
                self.outputs['image_header'] = d
            except Exception as e:
                d['ERROR'] = e.message
                self.outputs['image_header'] = d
                raise e




import numpy as np

from ... import Operation as opmod 
from ...Operation import Operation

class CSVToXYData(Operation):
    """
    Read a csv-formatted file as floats, 
    package into arrays of x values and y values.
    """

    def __init__(self):
        input_names = ['file_path']
        output_names = ['x','y','x_y']
        super(CSVToXYData, self).__init__(input_names, output_names)
        self.input_doc['file_path'] = 'path to .csv file'
        self.output_doc['x'] = 'numpy array built from first column of csv data'
        self.output_doc['y'] = 'numpy array built from second column of csv data'
        self.output_doc['x_y'] = 'n-by-2 numpy array with x and y values'

    def run(self):
        p = self.inputs['file_path']
        if p is None:
            return
        x = np.loadtxt(p, dtype=float, delimiter=',', usecols=[0])
        y = np.loadtxt(p, dtype=float, delimiter=',', usecols=[1])
        self.outputs['x'] = x
        self.outputs['y'] = y
        x_y = np.array([x,y]).T
        self.outputs['x_y'] = x_y 
        





import numpy as np
import pypif.obj as pifobj

from ... import Operation as opmod 
from ...Operation import Operation

class PifNPSynthExperiment(Operation):
    """
    Analyze a series of PIFs generated in a nanoparticle synthesis experiment
    and produce a master PIF that describes the overall experiment. 
    """

    def __init__(self):
        input_names = ['pifs','recipe_file','reaction_id']
        output_names = ['master_pif']
        super(PifNPSynthExperiment,self).__init__(input_names,output_names)
        self.input_doc['pifs'] = 'list of pypif.obj.ChemicalSystem objects '\
        'describing a full nanoparticle synthesis experiment. '
        self.input_doc['recipe_file'] = 'path to a YAML file '\
        'describing the synthesis recipe with a set of standard keys (TODO: document). '
        self.input_doc['reaction_id'] = 'string identifying the reaction: '\
        'because this is used as the ChemicalSystem.uid, it must be unique '\
        'among the uids in the data set where the PIF will be stored.'
        self.output_doc['master_pif'] = 'pypif.obj.ChemicalSystem object '\
        'describing the overall experiment. '

    def run(self):
        pifs = self.inputs['pifs']
        rcpfile = self.inputs['recipe_file']
        rxnid = self.inputs['reaction_id']
        csys = pifobj.ChemicalSystem()
        csys.uid = rxnid 
        # Subsystems for solution ingredients
        # TODO: Quantity information for subsystems
        #if t_T is not None:
        #    # Process time,temperature profile as a property
        #    it = (t_T[:,0] <= t_utc)
        #    t0 = t_T[0,0]
        #    t_T_current = np.array(t_T[it,:])
        #    t_T_current[:,0] = t_T_current[:,0] - t0
        #    csys.properties.append(self.time_feature_property(t_T_current,'temperature','degrees C'))
        #    # Grab the temperature at t_utc, else raise exception
        #    if t_utc in t_T[:,0]:
        #        temp_C = t_T[:,1][list(t_T[:,0]).index(t_utc)]
        #    else:
        #        raise ValueError('utc time {} not found in t_T data ({})'
        #        .format(t_utc,t_T[:,0]))
        # Package features-vs-time up to now as properties
        #if t_f is not None and not bflag:
        #    # Process time,features profile into properties
        #    #for fname in ['r0_form','sigma_form','I0_pre','I0_form','I_at_0']:
        #    bflags = np.array([fi['bad_data_flag'] for fi in t_f[:,1]],dtype=bool)
        #    pflags = np.array([fi['precursor_flag'] for fi in t_f[:,1]],dtype=bool)   
        #    sflags = np.array([fi['structure_flag'] for fi in t_f[:,1]],dtype=bool)   
        #    fflags = np.array([fi['form_flag'] for fi in t_f[:,1]],dtype=bool)   
        #    t0 = t_f[0,0]

        #    f_idx = np.where( (t_f[:,0]<=t_utc) & fflags & np.invert(sflags) & np.invert(bflags) )[0]
        #    if any(f_idx):
        #        t_r0 = np.array([[t_f[i][0],t_f[i][1]['r0_form']] for i in f_idx])
        #        t_r0[:,0] = t_r0[:,0] - t0 
        #        csys.properties.append(self.time_feature_property(t_r0,'nanoparticle mean radius vs. time','Angstrom'))
        #    
        #        t_sig = np.array([[t_f[i][0],t_f[i][1]['sigma_form']] for i in f_idx])
        #        t_sig[:,0] = t_sig[:,0] - t0 
        #        csys.properties.append(self.time_feature_property(t_sig,'fractional standard deviation of nanoparticle radius vs. time',''))

        #        t_I0form = np.array([[t_f[i][0],t_f[i][1]['I0_form']] for i in f_idx])
        #        t_I0form[:,0] = t_I0form[:,0] - t0 
        #        csys.properties.append(self.time_feature_property(t_I0form,'form factor scattering intensity prefactor vs. time','counts'))
        #        # Decide whether or not these are equilibrium features
        #        eqm_flag = None
        #        if t_utc-t0 == t_r0[-1,0] and t_r0.shape[1]>2:
        #            tvals = t_r0[-3:,0]
        #            r0vals = t_r0[-3:,1]
        #            p = np.polyfit(tvals,r0vals,2)
        #            dp = np.array([p[0]*2,p[1]])
        #            dr0dt = np.polyval(dp,t_r0[-1,0])
        #            eqm_flag = dr0dt < 0.01
        #        if eqm_flag is not None:
        #            csys.tags.append('chemical equilibrium: {}'.format(eqm_flag))
        #        else: 
        #            csys.tags.append('chemical equilibrium: unclear')
        #        
        #    p_idx = np.where( (t_f[:,0]<=t_utc) & pflags & np.invert(sflags) & np.invert(bflags) )[0]
        #    if any(p_idx):
        #        t_I0pre = np.array([[t_f[i][0],t_f[i][1]['I0_pre']] for i in p_idx])
        #        t_I0pre[:,0] = t_I0pre[:,0] - t0 
        #        csys.properties.append(self.time_feature_property(t_I0pre,'precursor scattering intensity prefactor vs. time','counts'))

        #    I0_idx = np.where( (t_f[:,0]<=t_utc) & np.invert(bflags) )[0]
        #    if any(I0_idx):
        #        t_I0 = np.array([[t_f[i][0],t_f[i][1]['I_at_0']] for i in I0_idx])
        #        t_I0[:,0] = t_I0[:,0] - t0 
        #        csys.properties.append(self.time_feature_property(t_I0,'scattered intensity at q=0 vs. time','counts'))

    def time_feature_property(self,t_f,fname,funits=''):
        pf = pifobj.Property()
        pf.name = fname 
        npts = t_f.shape[0]
        pf.scalars = [pifobj.Scalar(t_f[i,1]) for i in range(npts)]
        if funits:
            pf.units = funits 
        pf.conditions = []
        pf.conditions.append( pifobj.Value('reaction time',
                        [pifobj.Scalar(t_f[i,0]) for i in range(npts)],
                        None,None,None,'seconds') )
        return pf 






class TreeItem(object):
    """
    A structured container for indexing a TreeModel.
    A TreeItem keeps references to a parent TreeItem
    and a list of child TreeItems.
    It is labeled by a tag (TreeItem.tag)
    which must be unique across its sibling TreeItems.
    A root TreeItem should have None as its parent item.
    """

    def __init__(self,parent_itm,tag):
        super(TreeItem,self).__init__()
        self.parent = parent_itm 
        self.children = []      # list of child TreeItems
        self.flags = {}         # dict for any added functionality (e.g. toggles, etc) 
        self.tag = tag          # string tag for indexing and display
        #self.data = None       # TreeItem contains a single object as its data 

    def n_children(self):
        return len(self.children)

    def build_uri(self):
        """
        Return the TreeModel uri of this TreeItem
        by following its parents up to a root item. 
        """
        if self.parent is None:
            return ''
        else:
            uri = self.tag 
            itm = self.parent
            while itm.parent is not None:
                uri = itm.tag+'.'+uri
                itm = itm.parent
        return uri





from __future__ import print_function
import os
import string 
from collections import OrderedDict

class DictTree(object):
    """
    A tree as an ordered dictionary (root), 
    extended by embedding other objects 
    that are amenable to tree storage.
    Fetches items by a uri string that is a sequence 
    of dict keys, connected by '.'s.

    Child items (end nodes of the tree)
    can be anything.
    Parent items, in order to index their children,
    must be either lists, dicts, or objects implementing
    keys(), __getitem__(key) and __setitem__(key,value).
    """

    def __init__(self,data={}):
        super(DictTree,self).__init__()
        self._root = OrderedDict()
        if isinstance(data,dict):
            self._root = OrderedDict(data)
        self.bad_chars = string.punctuation 
        self.bad_chars = self.bad_chars.replace('_','')
        self.bad_chars = self.bad_chars.replace('-','')
        self.bad_chars = self.bad_chars.replace('.','')
        self.space_chars = [' ','\t','\n',os.linesep]
        #self._all_uris = []

    def __getitem__(self,uri):
        return self.get_from_uri(uri)

    def __setitem__(self,uri,val):
        self.set_uri(uri,val)

    def root_keys(self):
        return self._root.keys()

    def delete_uri(self,uri=''):
        """
        Delete the given uri, i.e., 
        remove the corresponding key, value pair 
        from the embedded dict.
        """
        try:
            itm = self._root
            if '.' in uri:
                itm = self.get_from_uri(uri[:uri.rfind('.')])
            path = uri.split('.')
            k = path[-1]
            if k:
                # Note- parent items must implement __getitem__
                del itm[k]
                #itm.pop(k)
        except Exception as ex:
            msg = str('\n[{}] Encountered an error while trying to delete uri {}: \n'
            .format(__name__,uri))
            ex.message = msg + ex.message
            raise ex

    def set_uri(self,uri='',val=None):
        """
        Set the data at the given uri to provided value val.
        """
        try:
            itm = self._root
            if '.' in uri:
                parent_uri = uri[:uri.rfind('.')]
                itm = self.get_from_uri(parent_uri)
            k = uri.split('.')[-1]
            # TODO: Is there a more graceful way to handle lists?
            if k:
                if isinstance(itm,list):
                    itm[int(k)] = val
                else:
                    # Note- parent items must implement __setitem__
                    itm[k] = val
                #if not uri in self._all_uris:
                #    self._all_uris.append(uri)
        except Exception as ex:
            msg = str('\n[{}] Encountered an error while trying to set uri {} to val {}: \n'
            .format(__name__,uri,val)) + ex.message
            raise KeyError(msg)

    def get_from_uri(self,uri=''):
        """
        Return the data stored at uri.
        Each data item in the lineage of the uri
        must implement __getitem__() with support for
        string-like keys, unless it is a list,
        in which case the key is cast as int(key)
        before using it as an index in the list.
        """
        try:
            path = uri.split('.')
            itm = self._root 
            while any(path):
                k = path.pop(0)
                # TODO: Is there a more graceful way to handle lists?
                if isinstance(itm,list):
                    itm = itm[int(k)]
                else:
                    # Note- parent items must implement __getitem__ and keys()
                    if k in itm.keys():
                        itm = itm[k]
                    else:
                        # this could be a dict with a key containing a '.'
                        found = False
                        while not found:
                            k = k+'.'+path.pop(0)                        
                            if k in itm.keys():
                                itm = itm[k]
                                found = True 
                            elif k+'.' in itm.keys():
                                itm = itm[k+'.']
                                found = True           
            return itm
        except Exception as ex:
            msg = str('[{}] Encountered an error while fetching uri {}: \n'
            .format(__name__,uri) + ex.message)
            raise KeyError(msg) 

    def is_uri_valid(self,uri):
        """
        Check for validity of a uri. 
        Uris may contain upper case letters, lower case letters, 
        numbers, dashes (-), and underscores (_). 
        Periods (.) are used as delimiters between tags in the uri.
        Any whitespace or any character in the string.punctuation library
        (other than -, _, or .) results in an invalid uri.
        """
        if not uri or any(map(lambda s: s in uri,self.space_chars))\
            or any(map(lambda s: s in uri,self.bad_chars)):
            return False 
        return True 

    def is_tag_valid(self,tag):
        """
        Check for validity of a tag.
        The conditions for a valid tag are the same as for a valid uri,
        except that a tag should not contain period (.) characters.
        """
        if not tag:
            return False
        elif '.' in tag:
            return False 
        else:
            return self.is_uri_valid(tag)

    def is_uri_unique(self,uri):
        """
        Check for uniqueness of a uri. 
        """
        return not self.contains_uri(uri)

    def uri_error_message(self,uri):
        """Provide a human-readable error message for bad uris."""
        if not uri:
            return 'error: entry is blank.'
        elif any(map(lambda s: s in uri,self.space_chars)):
            return 'error: \n"{}" contains whitespace.'.format(uri)
        elif any(map(lambda s: s in uri,self.bad_chars)):
            return 'error: \n"{}" contains special characters.'.format(uri)
        else:
            return 'error: \nunspecified error for "{}".'.format(uri)

    def tag_error_message(self,tag):
        """Provide a human-readable error message for bad tags."""
        if '.' in tag:
            return 'error: \n"{}" contains a period (.)\n'.format(tag)
        else:
            return self.uri_error_message(tag)

    #def contains_uri(self,uri):
    #    """Returns whether or not input uri points to an item in this tree."""
    #    return uri in self._all_uris
    def contains_uri(self,uri):
        """
        Check if the uri represents an item in this DictTree.
        """
        return uri in self.keys()

    def keys(self):
        return self.subkeys()

    def subkeys(self,root_uri=''):
        if not root_uri:
            itm = self._root
        else:
            itm = self.get_from_uri(root_uri)
        if isinstance(itm,list):
            rootks = list(map(lambda i: str(i),range(len(itm))))
        else:
            try:
                rootks = itm.keys()
            except AttributeError as ex:
                # non-parental nodes may have no keys()
                rootks = []
        if root_uri:
            prefix = root_uri + '.'
        else:
            prefix = root_uri
        subks = list(map(lambda s:prefix+s,rootks))
        for k in rootks:
            nextks = self.subkeys(prefix+k) 
            if any(nextks):
                subks = subks + nextks
        return subks

    def make_unique_uri(self,prefix):
        """
        Generate the next unique uri from prefix by appending '_x' to it, 
        where x is a minimal nonnegative integer.
        """
        suffix = 0
        gooduri = False
        #urilist = self._all_uris
        while not gooduri:
            testuri = prefix+'_{}'.format(suffix)
            if not self.contains_uri(testuri): 
                gooduri = True
            else:
                suffix += 1
        return testuri 

    def print_tree(self,root_uri='',rowprefix=''):
        """
        Print the content of the tree rooted at root_uri,
        with each row of the string preceded by rowprefix.
        """
        if root_uri:
            itm = self.get_from_uri(root_uri)
        else:
            itm = self._root
        if isinstance(itm,dict):
            tree_string = '\n'
            for k,x in itm.items():
                x_tree = self.print_tree(root_uri+'.'+k,rowprefix+'\t')
                tree_string = tree_string+rowprefix+'{}: {}\n'.format(k,x_tree)
        elif isinstance(itm,list):
            tree_string = '\n'
            for i,x in zip(range(len(itm)),itm):
                x_tree = self.print_tree(root_uri+'.'+str(i),rowprefix+'\t')
                tree_string = tree_string+rowprefix+'{}: {}\n'.format(i,x_tree)
        else:
            return '{}'.format(itm)
        return tree_string

    def print_tree(self,root_uri='',rowprefix=''):
        """
        Print the content of the tree rooted at root_uri,
        with each row of the string preceded by rowprefix.
        """
        if root_uri:
            itm = self.get_from_uri(root_uri)
        else:
            itm = self._root
        if isinstance(itm,dict):
            tree_string = '\n'
            for k,x in itm.items():
                x_tree = self.print_tree(root_uri+'.'+k,rowprefix+'\t')
                tree_string = tree_string+rowprefix+'{}: {}\n'.format(k,x_tree)
        elif isinstance(itm,list):
            tree_string = '\n'
            for i,x in zip(range(len(itm)),itm):
                x_tree = self.print_tree(root_uri+'.'+str(i),rowprefix+'\t')
                tree_string = tree_string+rowprefix+'{}: {}\n'.format(i,x_tree)
        else:
            return '{}'.format(itm)
        return tree_string

            



import datetime
import time

import tzlocal

from ... import Operation as opmod 
from ...Operation import Operation

class TimeTempFromHeader(Operation):
    """
    Get time and temperature from a detector output header file.
    Return string time, float time (utc in seconds), and float temperature.
    Time is assumed to be in the format Day Mon dd hh:mm:ss yyyy.
    """
    def __init__(self):
        input_names = ['image_header','time_key','temp_key']
        output_names = ['date_time','time','temperature']
        super(TimeTempFromHeader,self).__init__(input_names,output_names)        
        self.input_type['image_header'] = opmod.workflow_item
        self.input_doc['image_header'] = 'workflow uri of dict produced from detector output header file.'
        self.input_doc['time_key'] = 'key in image_header that refers to the time' 
        self.input_doc['temp_key'] = 'key in image_header that refers to the temperature' 
        self.output_doc['date_time'] = 'string representation of the time'
        self.output_doc['time'] = 'UTC time in seconds'
        self.output_doc['temperature'] = 'Temperature'

    def run(self):
        d = self.inputs['image_header']
        ktime = self.inputs['time_key']
        ktemp = self.inputs['temp_key']
        if d is None or ktime is None or ktemp is None:
            return
        time_str = str(d[ktime])
        temp = float(d[ktemp])
        # process the UTC time in seconds assuming %a %b %d %H:%M:%S %Y format
        # set local time zone for utc-awareness 
        tz = tzlocal.get_localzone()
        # use strptime to create a naive datetime object
        dt = datetime.datetime.strptime(time_str.strip(),"%a %b %d %H:%M:%S %Y")
        # add in timezone information to make a utc-aware datetime object
        dt_aware = datetime.datetime(dt.year,dt.month,dt.day,dt.hour,dt.minute,dt.second,dt.microsecond,tz)
        # interpret the time in UTC milliseconds
        t_utc = time.mktime(dt_aware.timetuple())
        self.outputs['date_time'] = time_str
        self.outputs['time'] = float(t_utc)
        self.outputs['temperature'] = temp




from collections import OrderedDict
import glob

from ...Operation import Operation
from ... import Operation as opmod 
from ... import optools

class BatchPostProcess(Operation):
    """
    Take the batch output (list of dicts) from a previously completed Batch,
    and use each dict to form inputs for the execution of a post-processing workflow.
    For each item to be taken from the previous batch, two keys are needed:
    one key indicates a previous batch workflow output, 
    and another indicates the corresponding current workflow input.
    """

    def __init__(self):
        input_names = ['batch_output','output_keys','workflow','input_keys']
        output_names = ['batch_inputs','batch_outputs']
        super(BatchPostProcess,self).__init__(input_names,output_names)
        self.input_doc['batch_output'] = 'list of dicts produced as batch output from another batch execution operation'
        self.input_doc['output_keys'] = 'list of keys for harvesting batch outputs (must be outputs of the previous batch workflow)'
        self.input_doc['workflow'] = 'a reference to the workflow to be executed in this batch'
        self.input_doc['input_keys'] = 'list of keys for setting workflow inputs, in corresponding order to output_keys'
        self.output_doc['batch_inputs'] = 'list of dicts of input_key:input_value for each of the input_keys'
        self.output_doc['batch_outputs'] = 'list of dicts of workflow outputs'
        self.input_type['workflow'] = opmod.entire_workflow
        self.input_type['batch_output'] = opmod.workflow_item
        
    def run(self):
        """
        Build a list of [uri:value] dicts to be used in the workflow.
        """
        batch_output = self.inputs['batch_output']
        out_keys = self.inputs['output_keys']
        inp_keys = self.inputs['input_keys']
        wf = self.inputs['workflow']
        if (wf is None or batch_output is None or out_keys is None or inp_keys is None):
            return
        self.outputs['batch_inputs'] = []
        self.outputs['batch_outputs'] = []
        n_batch = len(batch_output)
        for d_out in batch_output:
            inp_dict = OrderedDict() 
            for kout,kin in zip(out_keys,inp_keys):
                inp_dict[kin] = d_out[kout]
                wf.set_wf_input(kin,d_out[kout])
            wf.execute()
            self.outputs['batch_inputs'].append(inp_dict)
            self.outputs['batch_outputs'].append(wf.wf_outputs_dict())




from collections import OrderedDict

import numpy as np

from .. import Operation as opmod 
from ..Operation import Operation

class GetSAXSFlags(Operation):
    """
    Operation for retrieving SAXS population flags
    from a set of dicts read in from a previously saved YAML file 
    """

    def __init__(self):
        input_names = ['dicts','dict_key']
        output_names = ['flags']
        super(GetSAXSFlags, self).__init__(input_names, output_names)
        self.input_doc['dicts'] = 'dict of dicts read in from a YAML file where data labels were previously saved'
        self.input_doc['dict_key'] = 'The key to use to extract the flags from the input dicts'
        self.output_doc['flags'] = 'a dict with keys "bad_data", "precursor_scattering", '\
        '"form_factor_scattering", and "diffraction_peaks"' 

    def run(self):
        ds = self.inputs['dicts']
        k = self.inputs['dict_key']
        if ds is None or k is None:
            return
        f = OrderedDict()
        f['bad_data'] = ds['bad_data_flags'][k]
        f['precursor_scattering'] = ds['precursor_flags'][k]
        f['form_factor_scattering'] = ds['form_flags'][k]
        f['diffraction_peaks'] = ds['structure_flags'][k]

        self.outputs['flags'] = f 





import os

from pypif import pif
import pypif.obj as pifobj 

from ... import Operation as opmod 
from ...Operation import Operation
        
class SavePIFAsJSON(Operation):
    """
    Take a pypif.obj.System object and save it on the local filesystem in .json format
    """

    def __init__(self):
        input_names = ['pif','dirpath','filename']
        output_names = ['response']
        super(SavePIFAsJSON,self).__init__(input_names,output_names)
        self.input_doc['pif'] = 'A pypif.obj.System object or an array/list thereof'
        self.input_doc['dirpath'] = 'Path to system directory where .json file will be created' 
        self.input_doc['filename'] = 'Name of the .json file to be created. '\
        'The .json extension is appended automatically if not provided.' 
        self.input_type['pif'] = opmod.workflow_item
        self.input_type['filename'] = opmod.workflow_item

    def run(self):
        p = self.inputs['pif']        
        dp = self.inputs['dirpath']
        fn = self.inputs['filename']
        if dp is None or fn is None or p is None:
            return
        if not os.path.splitext(fn)[1] == 'json':
            fn = fn+'.json'
        json_file = os.path.join(dp,fn)
        pif.dump(p, open(json_file,'w'))




from collections import OrderedDict

from ...Operation import Operation
from ... import Operation as opmod 
from ... import optools

class RealtimeFromFiles(Operation):
    """
    Use file paths matching a regex to generate
    inputs for repeated execution of a workflow, 
    as the files arrive in a specified directory.
    Collects the outputs produced for each of the inputs.
    """

    def __init__(self):
        input_names = ['dir_path','regex','workflow','input_name','new_files_only','delay']
        output_names = ['realtime_inputs','realtime_outputs']
        super(RealtimeFromFiles,self).__init__(input_names,output_names)
        self.input_doc['dir_path'] = 'path to directory of input files'
        self.input_doc['regex'] = 'regular expression with wildcards, '\
            'used to filter or locate input files'
        self.input_doc['workflow'] = 'the Workflow to be executed'
        self.input_doc['input_name'] = 'name of the workflow input '\
            'where the file paths will be used'
        self.input_doc['new_files_only'] = 'if true, '\
            'ignore existing files and only process new arrivals'
        self.input_doc['delay'] = 'delay in milliseconds '\
            'before searching again when new files were not found '\
            'in the previous search'
        self.output_doc['realtime_inputs'] = 'list of dicts '\
            'containing [input_name:input_value] '\
            'for each file path used as workflow input'
        self.output_doc['realtime_outputs'] = 'list of dicts '\
            'containing [output_name:output_value] '\
            'for all of the workflow.outputs'
        self.input_type['workflow'] = opmod.entire_workflow
        self.inputs['new_files_only'] = True 
        self.inputs['delay'] = 100 
        
    def run(self):
        """
        This should create an iterator 
        whose next() gives a {uri:value} dict 
        built from the latest-arrived file 
        """
        dirpath = self.inputs['dir_path']
        rx = self.inputs['regex']
        wf = self.inputs['workflow'] 
        inpnm = self.inputs['input_name']
        if wf is None or not dirpath or not rx or not inpnm:
            return
        dly = self.inputs['delay']
        process_existing_files = not self.inputs['new_files_only']
        it = optools.FileSystemIterator(dirpath,rx,process_existing_files) 
        self.outputs['realtime_inputs'] = it
        self.outputs['realtime_outputs'] = [] 
        nx = 0 # total number of executions
        nd = 0 # number of consecutive delays
        keep_going = True
        while keep_going:
            p = it.next()
            if p is None:
                # delay
                time.sleep(float(dly)/1000.) 
                nd+=1
                if nd == 1000:
                    keep_going = False 
            else:
                nd = 0
                wf.set_wf_input(inpname,filename)
                nx+=1
                wf.execute()
                self.outputs['realtime_outputs'].append(wf.wf_outputs_dict())
                





import copy

from PySide import QtCore

from .QTreeModel import QTreeModel

class QTreeSelectionModel(QTreeModel):
    """
    QTreeSelectionModel extends QTreeModel 
    by using TreeItem.flags to handle tree item selection.
    """

    def __init__(self,flag_dict):
        super(QTreeSelectionModel,self).__init__(flag_dict)
        #self.flag_defaults = flag_defaults 

    def n_flags(self):
        return len(self.default_flags)

    def is_flagged(self,itm,flag_key):
        if flag_key in itm.flags.keys():
            return bool(itm.flags[flag_key])

    def children_flagged(self,itm,flag_key):
        if flag_key in itm.flags.keys():
            if itm.n_children() > 0:
                return any([self.children_flagged(c_itm,flag_key) for c_itm in itm.children])
            else:
                return bool(itm.flags[flag_key])

    def set_flagged(self,itm,flag_key,val):
        itm.flags[flag_key] = bool(val)

    def set_all_flagged(self,flag_key,val,itm=None):
        if itm is None:
            itm = self.root_item()
        self.set_flagged(itm,flag_key,val)
        for c_itm in itm.children:
            self.set_all_flagged(flag_key,val,c_itm)

    def get_flagged_idxs(self,flag_key,idx=None,val=True):
        if idx is None:
            idx = self.root_index()
        itm = self.get_from_index(idx) 
        flagged_idxs = []
        if self.is_flagged(itm,flag_key) == val:
            if not idx == self.root_index():
                flagged_idxs.append(idx)
        for c_row in range(itm.n_children()):
            c_idx = self.index(c_row,0,idx)
            flagged_idxs = flagged_idxs + self.get_flagged_idxs(flag_key,c_idx,val)
        return flagged_idxs

    def columnCount(self,parent):
        """
        Let QTreeSelectionModel have n_flags+1 columns:
        one for the TreeItem tag, the rest for flags 
        """
        return 1+self.n_flags() 

    def headerData(self,section,orientation,data_role):
        # note: section indicates row or column number, depending on orientation
        if (data_role == QtCore.Qt.DisplayRole and section == 0):
            return super(QTreeSelectionModel,self).headerData(section,orientation,data_role) 
        elif (data_role == QtCore.Qt.DisplayRole and section < self.n_flags()+1):
            return self.default_flags.keys()[section-1]
        else:
            return None

    def data(self,idx,data_role):
        if (not idx.isValid()):
            return None
        itm = idx.internalPointer()
        if idx.column() == 0:
            return super(QTreeSelectionModel,self).data(idx,data_role)
        elif data_role == QtCore.Qt.CheckStateRole: 
            return self.check_state(itm,self.default_flags.keys()[idx.column()-1])
        else:
            # Not column 0, not CheckStateRole: return None
            return None

    def check_state(self,itm,flag_key):
        if self.is_flagged(itm,flag_key):
            return QtCore.Qt.Checked
        elif self.children_flagged(itm,flag_key):
            return QtCore.Qt.PartiallyChecked
        else:
            return QtCore.Qt.Unchecked

    # Need flags to reflect checkability
    def flags(self, idx):
        if not idx.isValid():
            return None
        elif idx.column() == 0:
            return super(QTreeSelectionModel,self).flags(idx)
        else:
            return super(QTreeSelectionModel,self).flags(idx) | QtCore.Qt.ItemIsUserCheckable | QtCore.Qt.ItemIsTristate

    # Ui-editable QAbstractItemModel subclasses must implement setData(index,value[,role])
    def setData(self,idx,val,data_role):
        if idx.column() == 0:
            return super(QTreeSelectionModel,self).setData(index,val,data_role)
        elif data_role == QtCore.Qt.CheckStateRole:
            itm = self.get_from_index(idx)
            self.set_flagged(itm,self.default_flags.keys()[idx.column()-1],val)
            self.dataChanged.emit(idx,idx)
            return True
        else:
            return super(QTreeSelectionModel,self).setData(index,val,data_role)




"""This minimally enhances the paws.api module to interface with qt-based applications."""

from PySide import QtCore

from .. import api as pawsapi
from .QOpManager import QOpManager
from .QWfManager import QWfManager
from .QPluginManager import QPluginManager
from ..core import operations as ops

def start(app):
    """
    Instantiate and return a QPawsAPI object.
    Requires a valid QApplication as input. 

    paws.api.start() calls the QPawsAPI constructor.

    :returns: a QPawsAPI object
    :return type: paws.api.QPawsAPI 
    """
    return QPawsAPI(app)

class QPawsAPI(pawsapi.PawsAPI,QtCore.QObject):
    wfSelectionChanged = QtCore.Signal(str)

    def __init__(self,app):
        super(QPawsAPI,self).__init__()
        # Replace the PawsAPI TreeModels and workflow manager 
        # with the Qt-enabled variants.
        self._op_manager = QOpManager()
        self._plugin_manager = QPluginManager()
        self._wf_manager = QWfManager(app)
        self._wf_manager.plugin_manager = self._plugin_manager
        # TODO: load_cats and load_ops should happen outside the api.__init__
        # so that different api instances can have different operations loaded
        self._op_manager.load_cats(ops.cat_list) 
        self._op_manager.load_ops(ops.cat_op_list)
        self.app = app

    #def set_logmethod(self,lm):
    #    super(QPawsAPI,self).set_logmethod(lm)
    # TODO: figure out how to disconnect this signal
    # so that it is only connected to one message slot at a time
    #    self._wf_manager.emitMessage.disconnect()
    #    self._wf_manager.emitMessage.connect(lm)

    def get_op_from_index(self,idx):
        return self._op_manager.get_data_from_index(idx)

    def get_op_uri_from_index(self,idx):
        return self._op_manager.get_uri_of_index(idx)

    def get_plugin_from_index(self,idx):
        return self._plugin_manager.get_data_from_index(idx)

    def select_wf(self,wfname):
        super(QPawsAPI,self).select_wf(wfname)
        self.wfSelectionChanged.emit(wfname)
        
    def stop_wf(self,wfname):
        self._wf_manager.stop_wf(wfname)

    def run_wf(self,wfname,pool=None):
        """
        Run the workflow indicated by wfname.
        If optional threadpool is provided,
        the workflow attempts to run in that threadpool.
        """
        self._wf_manager.run_wf(wfname,pool)

    def is_wf_running(self,wfname):
        return self._wf_manager.wf_running[wfname]



from collections import OrderedDict

import yaml
import sklearn
from sklearn import preprocessing,linear_model
from collections import OrderedDict
import os

from ... import pawstools

class SaxsClassifier(object):
    """A container for a set of classifiers to be used on SAXS spectra"""

    def __init__(self,yml_file=None):
        if yml_file is None:
            p = os.path.abspath(__file__)
            # p = (pawsroot)/paws/core/tools/saxs/SaxsClassifier.py
            d = os.path.dirname(p)
            # d = (pawsroot)/paws/core/tools/saxs/
            yml_file = os.path.join(d,'modeling_data','scalers_and_models.yml')

        s_and_m_file = open(yml_file,'rb')
        s_and_m = yaml.load(s_and_m_file)

        sk_version = s_and_m['version']
        cur_version = list(map(int,sklearn.__version__.split('.')))
        major = cur_version[0]
        minor = cur_version[1]

        if (major != sk_version[0] or minor != sk_version[1]):
            version_str = ".".join(map(str,sk_version))
            msg = 'found scikit-learn v{}.{}.x. '\
                'SAXS classification models were built '\
                'with scikit-learn v{}. '\
                'Please try again with a matching version of scikit-learn.'\
                .format(major,minor,version_str) 
            raise RuntimeError(msg)

        scalers_dict = s_and_m['scalers'] # dict of scalers parametrs
        classifier_dict = s_and_m['models'] # dict of models parametrs

        model_param_bad_data = classifier_dict['bad_data']
        model_param_form = classifier_dict['form_factor_scattering']
        model_param_precur = classifier_dict['precursor_scattering']
        model_param_diff_peaks = classifier_dict['diffraction_peaks']

        scaler_param_bad_data = scalers_dict['bad_data']
        scaler_param_form = scalers_dict['form_factor_scattering']
        scaler_param_precur = scalers_dict['precursor_scattering']
        scaler_param_diff_peaks = scalers_dict['diffraction_peaks']

        # recreate the scalers
        scaler_bad_data = preprocessing.StandardScaler()
        self.set_param( scaler_bad_data, scaler_param_bad_data)

        scaler_form = preprocessing.StandardScaler()
        self.set_param( scaler_form, scaler_param_form)

        scaler_precur = preprocessing.StandardScaler()
        self.set_param( scaler_precur, scaler_param_precur)

        scaler_diff_peaks = preprocessing.StandardScaler()
        self.set_param( scaler_diff_peaks, scaler_param_diff_peaks)

        # recreate the models
        model_bad_data = linear_model.SGDClassifier()
        self.set_param( model_bad_data, model_param_bad_data)

        model_form = linear_model.SGDClassifier()
        self.set_param( model_form, model_param_form)

        model_precur = linear_model.SGDClassifier()
        self.set_param( model_precur, model_param_precur)

        model_diff_peaks = linear_model.SGDClassifier()
        self.set_param( model_diff_peaks, model_param_diff_peaks)

        # save the dicts of classifiers and scalers as outputs
        self.models = {'bad_data': model_bad_data,
                       'form_factor_scattering': model_form,
                       'precursor_scattering': model_precur,
                       'diffraction_peaks': model_diff_peaks}
        self.scalers = {'bad_data': scaler_bad_data,
                       'form_factor_scattering': scaler_form,
                       'precursor_scattering': scaler_precur,
                       'diffraction_peaks': scaler_diff_peaks}

    # helper function - to set parametrs for scalers and models
    def set_param(self, m_s, param):
        for k, v in param.items():
            if isinstance(v, list):
                setattr(m_s, k, np.array(v))
            else:
                setattr(m_s, k, v)

    def classify(self, sample_params):
        """Apply self.models and self.scalers to sample_params.

        Parameters
        ----------
        sample_params : array
            array of floats representing features of test sample

        Returns
        -------
        flags : dict
            dictionary of boolean flags indicating sample populations
        """ 
        flags = OrderedDict()
        x_bd = self.scalers['bad_data'].transform(sample_params)
        f_bd = self.models['bad_data'].predict(x_bd)[0]
        p_bd = self.models['bad_data'].predict_proba(x_bd)[0,int(f_bd)]
        flags['bad_data'] = (f_bd,p_bd)
        # NOTE: this is temporary, until new models have been built
        #flags['bad_data'] = (True,1.)
        if not f_bd: 
            for k in self.models.keys():
                if not k == 'bad_data':
                    xk = self.scalers[k].transform(sample_params)
                    fk = self.models[k].predict(xk)
                    pk = self.models[k].predict_proba(xk)[0,int(fk)]
                    flags[k] = (fk,pk)
        return flags

        #model_bad_data = clsfrs['bad_data']
        #model_form = clsfrs['form_factor_scattering']
        #model_precur = clsfrs['precursor_scattering']
        #model_diff_peaks = clsfrs['diffraction_peaks']

        #scaler_bad_data = sclrs['bad_data']
        #scaler_form = sclrs['form_factor_scattering']
        #scaler_precur = sclrs['precursor_scattering']
        #scaler_diff_peaks = sclrs['diffraction_peaks']

        ### Classify the spectrum
        # (1) Apply model to input

        # transform the ordered dictionary into an array of features:
        #bin_strengths = x['q_bin_strengths']
        #del x['q_bin_strengths']

        #edges = x['q_bin_edges']
        #del x['q_bin_edges']

        # we can use edges instead to indexes, but it could cause a problem
        # when not all samples have a specific edge
        #for i in range(len(edges)):
        #    x[str(edges[i])] = bin_strengths[i]

        # should we do it in profile_spectrum()?
        #for i in range(100):
        #    x[str(i)] = bin_strengths[i]

        # features for bad_data, precursor, and structure labels
        #features_analytical_and_60 = ['q_Imax', 'Imax_over_Imean', 'Imax_over_Ilowq',
        #'Imax_over_Ihighq', 'Imax_sharpness', 'low_q_ratio', 'high_q_ratio',
        #'log_fluctuation','0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12', '13',
        #'14', '15', '16', '17', '18', '19', '20', '21', '22', '23', '24', '25',
        #'26', '27', '28', '29', '30', '31', '32', '33', '34', '35', '36', '37',
        #'38', '39', '40', '41', '42', '43', '44', '45', '46', '47', '48', '49',
        #'50', '51', '52', '53', '54', '55', '56', '57', '58', '59' ]

        # features for form label
        #features60 = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12', '13',
        #'14', '15', '16', '17', '18', '19', '20', '21', '22', '23', '24', '25',
        #'26', '27', '28', '29', '30', '31', '32', '33', '34', '35', '36', '37',
        #'38', '39', '40', '41', '42', '43', '44', '45', '46', '47', '48', '49',
        #'50', '51', '52', '53', '54', '55', '56', '57', '58', '59']

        #input_features1 = [] # analytical and 60 bins
        #for item in features_analytical_and_60:
        #    input_features1.append(x[item])

        #input_features2 = [] # 60 bins
        #for item in features60:
        #    input_features2.append(x[item])

        #transformed_input_features = scaler_bad_data.transform([input_features1])
        #bad_data = model_bad_data.predict(transformed_input_features)[0] # [0] since we have only one sample
        #if bad_data:
        #    pr_bad_data = model_bad_data.predict_proba(transformed_input_features)[0, 1]
        #else:
        #    pr_bad_data = model_bad_data.predict_proba(transformed_input_features)[0, 0]

        #flags['bad_data'] = (bad_data, pr_bad_data) # label and propability to have this label

        #if bad_data == True:
        #    flags['precursor_scattering'] = (False, None)
        #    flags['form_factor_scattering'] = (False, None)
        #    flags['diffraction_peaks'] = (False, None)
        #else:
        #    #form label
        #    transformed_input_features = scaler_form.transform([input_features2])
        #    form = model_form.predict(transformed_input_features)[0]
        #    if form:
        #        pr_form = model_form.predict_proba(transformed_input_features)[0, 1]
        #    else:
        #        pr_form = model_form.predict_proba(transformed_input_features)[0, 0]
        #    flags['form_factor_scattering'] = (form, pr_form) # label and propability to have this label

        #    # precursor label
        #    transformed_input_features = scaler_precur.transform([input_features1])
        #    prec = model_precur.predict(transformed_input_features)[0]
        #    if prec:
        #        pr_prec = model_precur.predict_proba(transformed_input_features)[0, 1]
        #    else:
        #        pr_prec = model_precur.predict_proba(transformed_input_features)[0, 0]
        #    flags['precursor_scattering'] = (prec, pr_prec)

        #    # difraction peaks label
        #    transformed_input_features = scaler_diff_peaks.transform([input_features1])
        #    picks = model_diff_peaks.predict(transformed_input_features)[0]
        #    if picks:
        #        pr_picks = model_diff_peaks.predict_proba(transformed_input_features)[0, 1]
        #    else:
        #        pr_picks = model_diff_peaks.predict_proba(transformed_input_features)[0, 0]
        #    flags['diffraction_peaks'] = (picks, pr_picks)

        # problems for later:
        # if flags['form_factor_scattering']:
        #     # classify the form factor based on inputs  
        #     flags['form_factor_id'] = ''     
        #
        # if flags['diffraction_peaks']:
        #     # classify the space group based on inputs
        #     flags['space_group_id'] = ''
        #

        #self.outputs['population_flags'] = flags 







import numpy as np
from collections import OrderedDict

from ... import Operation as opmod 
from ...Operation import Operation
from ....tools.saxs import saxs_fit 

class SpectrumProfiler(Operation):
    """
    This operation profiles a SAXS spectrum (I(q) vs. q)
    by taking various scalar quantities from the data.

    Outputs a dictionary of the results.

    This Operation is somewhat robust for noisy data,
    but any preprocessing (background subtraction, smoothing, or other cleaning)
    should be performed beforehand. 
    """

    def __init__(self):
        input_names = ['q_I', 'dI']
        output_names = ['features']
        super(SpectrumProfiler, self).__init__(input_names, output_names)
        self.input_doc['q_I'] = 'n-by-2 array of q(1/Angstrom) versus I(arb).'
        self.output_doc['features'] = 'dict profiling the input spectrum. '\
        'See the documentation of paws.core.tools.saxs.saxs_fit.profile_spectrum().'
        self.input_type['q_I'] = opmod.workflow_item
        self.input_type['dI'] = opmod.no_input

    def run(self):
        q_I = self.inputs['q_I']
        d_prof = saxs_fit.profile_spectrum(q_I)
        self.outputs['features'] = d_prof




from ... import Operation as opmod 
from ...Operation import Operation

class ShipJSON(Operation):
    """
    Take a .json file containing a pif or array of pifs, ship it to a Citrination data set.    
    """

    def __init__(self):
        input_names = ['json_path','client','dsid','ship_flag']
        output_names = ['response']
        super(ShipJSON,self).__init__(input_names,output_names)
        self.input_doc['json_path'] = 'Filesystem path where the json of the pif is saved' 
        self.input_doc['client'] = 'A working Citrination client' 
        self.input_doc['dsid'] = 'Data set ID where the pif will be stored on Citrination' 
        self.input_doc['ship_flag'] = 'Flag for actually shipping the pif' 
        self.output_doc['response'] = 'The Citrination server response to the shipment'
        self.input_type['client'] = opmod.workflow_item
        self.input_type['dsid'] = opmod.workflow_item
        self.inputs['ship_flag'] = False

    def run(self):
        json_path = self.inputs['json_path']
        cl = self.inputs['client'] 
        dsid = self.inputs['dsid'] 
        ship_flag = self.inputs['ship_flag']
        try:
            if ship_flag:
                r = cl.upload_file(json_path,data_set_id = dsid)
            else:
                r = 'dry run: no shipment occurred. json path: {}'.format(json_path)
        except Exception as ex:
            r = 'An error occurred while shipping. Error message: {}'.format(ex.message)
        self.outputs['response'] = r




import numpy as np
from matplotlib.figure import Figure
from matplotlib.backends.backend_qt4agg import FigureCanvasQTAgg as FigCanvas
#import matplotlib
#from matplotlib.backends import qt_compat

def array_plot_2d(data_in):
    return mpl_array_plot_2d(data_in)

def array_plot_1d(data_in):
    return mpl_array_plot_1d(data_in)

def mpl_array_plot_1d(data_in):
    fig = Figure(figsize=(100,100))
    axes = fig.add_subplot(111)
    dims = np.shape(data_in)
    if len(dims) > 1:
        if dims[1]==2:
            axes.plot(data_in[:,0],data_in[:,1])
        elif dims[0]==2: 
            axes.plot(data_in[0,:],data_in[1,:])
    else:
        axes.plot(data_in)
    return FigCanvas(fig)

def mpl_array_plot_2d(data_in):
    fig = Figure(figsize=(100,100))
    axes = fig.add_subplot(111)
    axes.contour(data_in)
    # FigCanvas is a subclass of QWidget
    return FigCanvas(fig)

def plot_mpl_fig(fig_in):
    return FigCanvas(fig_in)





from PySide import QtCore

class ListModel(QtCore.QAbstractListModel):
    """
    Class for list management with a QAbstractListModel.
    Implements required virtual methods rowCount() and data().
    Resizeable ListModels must implement insertRows(), removeRows().
    If a nicely labeled header is desired, implement headerData().
    """

    def __init__(self,input_list=[],parent=None):
        super(ListModel,self).__init__(parent)
        self._list_data = []
        self._enabled = []
        if input_list is not None:
            for thing in input_list:
                self.append_item(thing)
        
    def append_item(self,thing):
        ins_row = self.rowCount()
        self.beginInsertRows(QtCore.QModelIndex(),ins_row,ins_row)
        self._list_data.insert(ins_row,thing) 
        self.endInsertRows()
        self._enabled.insert(ins_row,True)
        idx = self.index(ins_row,0,QtCore.QModelIndex())

    def remove_item(self,row):
        self.beginRemoveRows(QtCore.QModelIndex(),row,row)
        self._list_data.pop(row) 
        self.endRemoveRows()
        self._enabled.pop(row) 

    #def list_data_changed(self):
    #    for r in range(len(self._list_data)):
    #        idx = self.index(r,0,QtCore.QModelIndex())
    #        self.dataChanged.emit(idx,idx)

    def set_enabled(self,row):
        self._enabled[row] = True

    def set_disabled(self,row):
        self._enabled[row] = False

    def list_data(self):
        return self._list_data

    def flags(self,idx):
        if self._enabled[idx.row()]:
            return QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable
        else:
            return QtCore.Qt.NoItemFlags

    def n_items(self):
        return self.rowCount()

    def rowCount(self,parent=QtCore.QModelIndex()):
        return len(self._list_data)

    def columnCount(self,parent=QtCore.QModelIndex()):
        return 1

    def data(self,idx,data_role):
        if not idx.isValid():
            return None
        elif (data_role == QtCore.Qt.DisplayRole
        or data_role == QtCore.Qt.ToolTipRole 
        or data_role == QtCore.Qt.StatusTipRole
        or data_role == QtCore.Qt.WhatsThisRole):
            return str(self._list_data[idx.row()])
        else:
            return None
        #    print 'data at row {}: {}'.format(idx.row(),self._list_data[idx.row()])
        #    print 'DATA IS NONE'

    def get_item(self,idx):
        return self._list_data[idx.row()]

    def insertRows(self,row,count):
        self.beginInsertRows(QtCore.QModelIndex(),row,row+count-1)
        for j in range(row,row+count):
            self._list_data.insert(j,None)
        self.endInsertRows()

    def removeRows(self, row, count, parent=QtCore.QModelIndex()):
        if all([r in range(self.rowCount(parent)) for r in range(row,count)]):
            self.beginRemoveRows(parent,row,row+count-1)
            for j in range(row,row+count)[::-1]:
                self._list_data.pop(j)
            self.endRemoveRows()
            return True
        else:
            return False

    def headerData(self,section,orientation,data_role):
        if (data_role == QtCore.Qt.DisplayRole and section == 0):
            return "list: {} item(s)".format(self.rowCount(QtCore.QModelIndex()))
        else:
            return None

class PluginListModel(ListModel):
    """Just a ListModel with overloaded headerData"""

    def __init__(self,input_list=[],parent=None):
        super(PluginListModel,self).__init__(input_list,parent)

    def headerData(self,section,orientation,data_role):
        if (data_role == QtCore.Qt.DisplayRole and section == 0):
            return "{} plugin(s) available".format(self.rowCount(QtCore.QModelIndex()))
        else:
            return None






import os

from pypif import pif
import pypif.obj as pifobj 

from ... import Operation as opmod 
from ...Operation import Operation
        
class ShipToDataSet(Operation):
    """
    Take a pypif.obj.System object and ship it to a given Citrination data set.    
    """

    def __init__(self):
        input_names = ['pif','client','dsid','json_dirpath','json_filename','keep_json','ship_flag']
        output_names = ['response']
        super(ShipToDataSet,self).__init__(input_names,output_names)
        self.input_doc['pif'] = 'A pypif.obj.System object or an array/list thereof'
        self.input_doc['client'] = 'A running Citrination client' 
        self.input_doc['dsid'] = 'Data set ID where the pif record(s) will be stored on Citrination' 
        self.input_doc['json_dirpath'] = 'Filesystem path where a json file of the pif(s) will be saved' 
        self.input_doc['json_filename'] = 'Name of the .json file where the pif(s) will be saved' 
        self.input_doc['keep_json'] = 'Flag for whether or not to keep the json file' 
        self.input_doc['ship_flag'] = 'Flag for shipping the pif- set to False for a dry run' 
        self.output_doc['response'] = 'The Citrination server response to the shipment'
        self.input_type['pif'] = opmod.workflow_item
        self.input_type['client'] = opmod.plugin_item
        self.inputs['ship_flag'] = False
        self.inputs['keep_json'] = False

    def run(self):
        cl = self.inputs['client'] 
        dsid = self.inputs['dsid'] 
        p = self.inputs['pif']        
        json_dir = self.inputs['json_dirpath']
        json_file = self.inputs['json_filename']
        if cl is None or dsid is None or p is None \
        or json_dir is None or json_file is None:
            return
        if not os.path.splitext(json_file)[1] == 'json':
            json_file = json_file+'.json'
        json_file = os.path.join(json_dir,json_file)

        json_flag = self.inputs['keep_json']
        ship_flag = self.inputs['ship_flag']
        try:
            # make p an array of pifs to get a big json that has all records
            pif.dump(p, open(json_file,'w'))
            if ship_flag:
                r = cl.upload_file(json_file,dataset_id = dsid)
            else:
                r = 'dry run: no shipment occurred. pif object: {}'.format(pif.dumps(p))
            if not json_flag:
                os.remove(json_file) 
        except Exception as ex:
            r = 'An error occurred while shipping. Error message: {}'.format(ex.message)
        self.outputs['response'] = r





from collections import OrderedDict

from ... import Operation as opmod 
from ...Operation import Operation

class ReadNPSynthRecipe(Operation):
    """
    Read in a text file describing nanoparticle synthesis parameters.
    Package the recipe description in a dict. 
    """
    
    def __init__(self):
        input_names = ['file_path']
        output_names = ['recipe_dict']
        super(ReadNPSynthRecipe,self).__init__(input_names,output_names)
        self.input_doc['file_path'] = 'plain text file describing a synthesis recipe'
        self.output_doc['recipe_dict'] = 'dict describing the synthesis recipe'

    def run(self):
        fpath = self.inputs['file_path']
        if fpath is None:
            return
        rdict = OrderedDict()
        ln = 0
        for line in open(fpath,'r'):
            ln += 1
            rdict[str(ln)] = line
        self.outputs['recipe_dict'] = rdict 




import os.path

from ..Operation import Operation
from .. import Operation as opmod 

class BuildFilePath(Operation):
    """
    This operation helps to build file paths from workflow data.
    It takes a directory (full path), a filename, and an extension.
    The filename can optionally have a prefix or suffix inserted,
    to help with iteration of batches of files with similar names.
    """

    def __init__(self):
        input_names = ['dir_path','prefix','filename','suffix','ext']
        output_names = ['filename','file_path']
        super(BuildFilePath, self).__init__(input_names, output_names)
        self.input_doc['dir_path'] = 'filesystem path pointing to the directory containing the file- a trailing slash is optional'
        self.input_doc['prefix'] = 'any text to prepend to filename (prefix comes after dir_path, before filename)'
        self.input_doc['filename'] = 'name of the file, excluding any path, extension, prefix, or suffix'
        self.input_doc['suffix'] = 'any text to append to filename (comes after filename, before ext)'
        self.input_doc['ext'] = 'extension for the file- the . is optional'
        self.input_type['filename'] = opmod.workflow_item
        self.inputs['prefix'] = ''
        self.inputs['suffix'] = ''
        self.inputs['ext'] = ''
        self.output_doc['filename'] = 'filename will be <prefix><filename><suffix>' 
        self.output_doc['file_path'] = 'file_path will be <path><prefix><filename><suffix>.<ext>' 

    def run(self):
        p = self.inputs['dir_path']
        fn = self.inputs['filename']
        if p is None or fn is None:
            return
        ext = self.inputs['ext']
        if not ext[0] == '.' and not ext == '':
            ext = '.'+ext
        #if not p[-1] == '/':
        #    p = p + '/'
        pf = self.inputs['prefix']
        sf = self.inputs['suffix']
        self.outputs['filename'] = str(pf+fn+sf) 
        full_filename = self.outputs['filename']+ext
        #full_filename = os.path.join(self.outputs['filename'],ext)
        self.outputs['file_path'] = os.path.join(p,full_filename)




import pyqtgraph as pg

def array_plot_2d(data_in):
    return pqg_array_plot_2d(data_in)

def array_plot_1d(data_in):
    return pqg_array_plot_1d(data_in)

def pqg_array_plot_2d(data_in):
    widg = pg.ImageView()
    widg.setImage(data_in)
    return widg 

def pqg_array_plot_1d(data_in):
    widg = pg.PlotWidget()
    widg.getPlotItem().plot(data_in)
    return widg 

def plot_mpl_fig(fig_in):
    return FigCanvas(fig_in)




from ..Operation import Operation

class Identity(Operation):
    """An Operation testing class, loads its input into its output"""

    def __init__(self):
        input_names = ['data']
        output_names = ['data']
        super(Identity,self).__init__(input_names,output_names) 
        self.input_doc['data'] = 'this can actually be anything'
        self.output_doc['data'] = 'this ends up being whatever the input was'
        
    def run(self):
        self.outputs['data'] = self.inputs['data']



from __future__ import print_function
from collections import OrderedDict
import copy
import traceback
import time

from .. import operations as ops
from .Workflow import Workflow
from ..operations import Operation as opmod
from ..operations.Operation import Operation#, Batch, Realtime        
from ..operations import optools
from .. import pawstools

class WfManager(object):
    """
    Manager for paws Workflows. 
    Stores a list of Workflow objects, 
    performs operations on them.
    Keeps a reference to a PluginManager
    for access to PawsPlugins.
    """

    def __init__(self):
        super(WfManager,self).__init__()
        self.workflows = OrderedDict() 
        self.logmethod = print 
        self.plugin_manager = None

    def get_op(self,wfname,op_tag):
        return self.workflows[wfname].get_data_from_uri(op_tag)

    def n_wf(self):
        return len(self.workflows)

    def add_wf(self,wfname):
        """
        Add a workflow to self.workflows, with key specified by wfname.
        If wfname is not unique (i.e. a workflow with that name already exists),
        this method will overwrite the existing workflow with a new one.
        """
        wf = Workflow()
        if not wf.is_tag_valid(wfname): 
            raise pawstools.WfNameError(wf.tag_error_message(wfname))
        wf.message_callback = self.logmethod
        self.workflows[wfname] = wf

    def run_wf(self,wfname):
        """
        Execute the workflow indicated by input wfname
        """
        wf = self.workflows[wfname]
        self.logmethod('preparing workflow {} for execution'.format(wfname))
        stk,diag = wf.execution_stack()
        self.prepare_wf(wf,stk)
        wf.execute()
        self.logmethod('execution finished')

    def prepare_wf(self,wf,stk):
        """
        For all of the operations in stack stk,
        load all inputs that are not workflow items. 
        """
        for lst in stk:
            for op_tag in lst:
                op = wf.get_data_from_uri(op_tag)
                for inpname,il in op.input_locator.items():
                    if il.tp not in [opmod.runtime_type,opmod.workflow_item]:
                        # runtime inputs should be set directly, without using il.val.
                        # this is because il.val gets serialized in wf.wf_setup_dict().
                        # workflow_item inputs should be set later, during execution.
                        # the no_input case ends up setting the input to None
                        #op.inputs[inpname] = self.locate_input(il)
                        wf.set_op_item(op_tag,'inputs.'+inpname,self.locate_input(il))

    def locate_input(self,il):
        """
        Return the data pointed to by a given InputLocator object.
        """
        if il.tp == opmod.no_input or il.val is None:
            return None
        elif il.tp == opmod.basic_type:
            return il.val
        elif il.tp == opmod.entire_workflow:
            wf = self.workflows[il.val]
            stk,diag = wf.execution_stack()
            self.prepare_wf(wf,stk)
            return wf
            #return self.workflows[il.val]
        elif il.tp == opmod.plugin_item:
            if isinstance(il.val,list):
                return [self.plugin_manager.get_data_from_uri(v) for v in il.val]
            else:
                return self.plugin_manager.get_data_from_uri(il.val)
        # changed: let basic_type inputs be loaded directly,
        # without using InputLocators.
        #elif il.tp == opmod.basic_type:
        #    return il.val

    # TODO: the following
    def check_wf(self,wf):
        """
        Check the dependencies of the workflow.
        Ensure that all loaded operations have inputs that make sense.
        Return a status code and message for each of the Operations.
        """
        pass

    def uri_to_embedded_dict(self,uri,data=None):
        path = uri.split('.')
        endtag = path[-1]
        d = OrderedDict()
        d[endtag] = data
        for tag in path[:-1][::-1]:
            parent_d = OrderedDict()
            parent_d[tag] = d
            d = parent_d
        return d

    def update_embedded_dict(self,d,d_new):
        for k,v in d_new.items():
            if k in d.keys():
                if isinstance(d[k],dict) and isinstance(d_new[k],dict):
                    # embedded dicts: recurse
                    d[k] = self.update_embedded_dict(d[k],d_new[k])
                else:
                    # existing d[k] is not dict, and d_new[k] is dict: replace  
                    d[k] = d_new[k]
            else:
                # d[k] does not exist: insert
                d[k] = v
        return d

    def load_from_dict(self,wfname,wf_spec,op_manager):
        """
        Create a workflow with name wfname.
        If wfname is not unique, self.workflows[wfname] is overwritten.
        Input dict wf_spec specifies Workflow setup,
        including all operations, Workflow.inputs, and Workflow.outputs.
        """
        self.add_wf(wfname)
        wfins = wf_spec.pop('WORKFLOW_INPUTS')
        wfouts = wf_spec.pop('WORKFLOW_OUTPUTS')
        opflags = wf_spec.pop('OP_ENABLE_FLAGS')
        for inpname,inpval in wfins.items():
            self.workflows[wfname].connect_wf_input(inpname,inpval)
        for outname,outval in wfouts.items():
            self.workflows[wfname].connect_wf_output(outname,outval)
        for op_tag, op_setup in wf_spec.items():
            self.workflows[wfname].add_op(op_tag,\
            self.workflows[wfname].build_op_from_dict(op_setup,op_manager))
            self.workflows[wfname].set_op_enabled(op_tag,opflags[op_tag])




from __future__ import print_function

from pypif import pif
from citrination_client import CitrinationClient 

from .. import pawstools
from .PawsPlugin import PawsPlugin
from ..operations import Operation as opmod

class CitrinationPlugin(PawsPlugin):
    """
    Wrapper contains a Citrination client and
    implements the PawsPlugin abc interface.
    """

    def __init__(self):
        input_names = ['address','api_key_file']
        super(CitrinationPlugin,self).__init__(input_names)
        self.input_doc['address'] = 'web address of citrination instance'
        self.input_doc['api_key_file'] = 'path to a file in the local filesystem containing a valid citrination api key'
        self.inputs['address'] = 'http://citrination.com' 
        self.ctn_client = None
        self.return_codes = {} 

    def start(self):
        self.address = self.inputs['address'] 
        f = open(self.inputs['api_key_file'],'r')
        self.api_key = str(f.readline()).strip()
        f.close()
        self.ctn_client = CitrinationClient(api_key = self.api_key, site = self.address)

    def stop(self):
        pass

    def content(self): 
        return {'client':self.ctn_client,'inputs':self.inputs}

    def description(self):
        desc = str('Citrination API Client Plugin for Paws: '
            + 'This is a container for the Citrination Client module. '
            + 'The Citrination Client connects to a Citrination instance '
            + 'and exposes some parts of the Citrination API. '
            + 'Startup requires the web address of a Citrination instance '
            + 'and an API key that provides access to that instance.')
        return desc

    def ship_dataset(self,pifs):
        # Create the data set
        response = self.ctn_client.create_data_set()
        dsid = response.json()['id']
        # TODO: Note that the entire data set can be one json,
        # of an array of pif records, and this will lead to a faster upload.
        for p in pifs:
            try:
                json_file = pawstools.scratchdir+'/'+p.uid+'.json'
                pif.dump(p, open(json_file,'w'))
                #print 'add DATA SET {} to tags'.format(dsid)
                #p.tags.append('DATA SET {}'.format(dsid))
                #print 'dump {} to data set {}'.format(json_file,dsid)
                cl.upload_file(json_file,data_set_id = dsid)
                #print 'NOT SHIPPING {} (this is a test)'.format(json_file)
                self.return_codes[p.uid]=1
                # delete dataset json
                #print 'deleting file {}'.format(json_file)
                os.remove(json_file) 
            except:
                # TODO: Pass along some return code from the server?
                self.return_codes[p.uid]=-1





from collections import OrderedDict

from PySide import QtCore

from .QTreeSelectionModel import QTreeSelectionModel
from ..core.plugins.PawsPlugin import PawsPlugin
from ..core.operations import Operation as opmod
from ..core.plugins.PluginManager import PluginManager

class QPluginManager(PluginManager,QTreeSelectionModel):
    """
    A Qt Signal-slot manager for a TreeModel PluginManager.
    Takes a reference to a PluginManager in the constructor.
    The QPluginManager works mostly by
    calling on the methods of the PluginManager.
    """

    def __init__(self):
        super(QPluginManager,self).__init__()

    def headerData(self,section,orientation,data_role):
        if (data_role == QtCore.Qt.DisplayRole and section == 0):
            return "{} plugin(s) loaded".format(self.n_plugins())
        else:
            return super(QPluginManager,self).headerData(section,orientation,data_role)    

    @QtCore.Slot(str)
    def update_plugin(self,pgin_name):
        self.set_item(pgin_name,self._tree.get_data_from_uri(pgin_name))





import numpy as np

from ... import Operation as opmod 
from ...Operation import Operation

class Rotation(Operation):
    """Rotate an array by 90, 180, or 270 degrees."""

    def __init__(self):
        input_names = ['image_data','rotation_deg']
        output_names = ['image_data']
        super(Rotation,self).__init__(input_names,output_names)        
        self.input_doc['image_data'] = '2d array representing intensity for each pixel'
        self.input_doc['rotation_deg'] = 'rotation in degrees counter-clockwise: '\
            'must be one of 90, 180, or 270'
        self.output_doc['image_data'] = '2d array representing rotated image'
        self.inputs['rotation_deg'] = 90 

    def run(self):
        """Rotate self.inputs['image_data'] and save as self.outputs['image_data']"""
        img = self.inputs['image_data']
        if img is None:
            return
        rot_deg = int(self.inputs['rotation_deg'])
        if rot_deg==90:
            img_rot = np.rot90(img)
        elif rot_deg==180:
            img_rot = np.rot90(np.rot90(img))
        elif rot_deg==270:
            img_rot = np.rot90(np.rot90(np.rot90(img)))
        else:
            msg = '[{}] expected rot_deg = 90, 180, or 270, got {}'.format(__name__,rot_deg)
            raise ValueError(msg)
        # save results to self.outputs
        self.outputs['image_data'] = img_rot



import numpy as np

from .. import Operation as opmod 
from ..Operation import Operation

class Window(Operation):
    """
    Window an n-by-2 array x_y 
    such that x is bounded by specified limits 
    """

    def __init__(self):
        input_names = ['x_y','x_min','x_max']
        output_names = ['x_y_window']
        super(Window,self).__init__(input_names,output_names)        
        self.input_type['x_y'] = opmod.workflow_item
        self.input_doc['x_y'] = 'n-by-2 array of x and y values'
        self.input_doc['x_min'] = 'inclusive minimum x value of output'
        self.input_doc['x_max'] = 'inclusive maximum x value of output'
        self.output_doc['x_y_window'] = 'n-by-2 array with x, y pairs for x_min <= x <= x_max'

    def run(self):
        x_y = self.inputs['x_y']
        x_min = self.inputs['x_min']
        x_max = self.inputs['x_max']
        idx_keep = ((x_y[:,0] >= x_min) & (x_y[:,0] <= x_max))
        self.outputs['x_y_window'] = x_y[idx_keep,:] 




import numpy as np
import pypif.obj as pifobj

from ... import Operation as opmod 
from ...Operation import Operation
from ....tools.saxs import saxs_fit 

class PifNPSolutionSAXS(Operation):
    """
    Package SAXS results from a nanoparticle solution into a pypif.obj.ChemicalSystem record.
    """

    def __init__(self):
        input_names = ['uid_prefix','date_time','t_utc','q_I','temperature','features']
        output_names = ['pif']
        super(PifNPSolutionSAXS,self).__init__(input_names,output_names)
        self.input_doc['uid_prefix'] = 'text string to prepend to pif uid (pif uid = uid_prefix+t_utc'
        self.input_doc['date_time'] = 'string date/time from measurement header file, used for pif record tags'
        self.input_doc['t_utc'] = 'time in seconds utc'
        self.input_doc['q_I'] = 'n-by-2 array of q values and corresponding intensities for saxs spectrum'
        self.input_doc['temperature'] = 'temperature of the system in degrees Celsius'
        self.input_doc['features'] = 'dict of features extracted from the measured spectrum.'
        self.output_doc['pif'] = 'pif object representing the input data'
        self.input_type['date_time'] = opmod.workflow_item
        self.input_type['t_utc'] = opmod.workflow_item
        self.input_type['q_I'] = opmod.workflow_item
        self.input_type['temperature'] = opmod.workflow_item
        self.input_type['features'] = opmod.workflow_item
        # all inputs default to none: an empty pif should be produced in this case.
        self.inputs['uid_prefix'] = None
        self.inputs['date_time'] = None
        self.inputs['t_utc'] = None
        self.inputs['q_I'] = None
        self.inputs['temperature'] = None
        self.inputs['features'] = None

    def run(self):
        uid_pre = self.inputs['uid_prefix']
        t_str = self.inputs['date_time']
        t_utc = self.inputs['t_utc']
        uid_full = 'tmp'
        if uid_pre is not None:
            uid_full = uid_pre
        if t_utc is not None:
            uid_full = uid_full+'_'+str(int(t_utc))
        temp_C = self.inputs['temperature']
        q_I = self.inputs['q_I']
        ftrs = self.inputs['features']

        # ChemicalSystem record construction begins here
        csys = pifobj.ChemicalSystem()
        csys.uid = uid_full
        csys.properties = []
        csys.tags = []
        if uid_pre is not None:
            csys.tags.append('reaction id: '+uid_pre)
        if t_str is not None:
            csys.tags.append('date: '+t_str)
        if t_utc is not None:
            csys.tags.append('utc: '+str(int(t_utc)))

        # TODO: include the scattering equation in this record somehow
        if ftrs is not None:
            if ftrs['bad_data']:
                csys.tags.append('unidentified scattering or bad data')
            if ftrs['diffraction_peaks']:
                csys.tags.append('diffraction peaks')
            if ftrs['form_factor_scattering']:
                csys.tags.append('form factor scattering')
                if not ftrs['diffraction_peaks'] and not ftrs['bad_data']:
                    csys.properties.append(self.feature_property(None,'nanoparticle mean radius','Angstrom'))
                    csys.properties.append(self.feature_property(None,'fractional standard deviation of nanoparticle radius',''))
                    csys.properties.append(self.feature_property(None,'form factor scattering intensity prefactor','counts'))
            if ftrs['precursor_scattering']:
                csys.tags.append('precursor scattering')
                if not ftrs['diffraction_peaks'] and not ftrs['bad_data']:
                    csys.properties.append(self.feature_property(None,'precursor effective radius','Angstrom'))
                    csys.properties.append(self.feature_property(None,'precursor scattering intensity prefactor','counts'))

        if q_I is not None:
            # Process measured q_I into a property
            pI = self.q_I_property(q_I)
            if temp_C is not None:
                pI.conditions.append(pifobj.Value('temperature',[pifobj.Scalar(temp_C)],None,None,None,'degrees Celsius'))
            pI.name = 'SAXS intensity' 
            csys.properties.append(pI)
            if not ftrs['diffraction_peaks'] and not ftrs['bad_data']:
                csys.properties.append(self.feature_property(None,'scattered intensity at q=0','counts'))
                # Compute the saxs spectrum, package it
                qI_computed = saxs_fit.compute_saxs(q_I[:,0],ftrs)
                pI_computed = self.q_I_property(np.array( [[q_I[i,0],qI_computed[i]] for i in range(len(qI_computed))] ))
                pI_computed.name = 'computed SAXS intensity'
                csys.properties.append(pI_computed)
        self.outputs['pif'] = csys

    def feature_property(self,fval,fname,funits=''):
        pf = pifobj.Property()
        pf.name = fname
        pf.scalars = [pifobj.Scalar(fval)]
        if funits:
            pf.units = funits
        return pf

    def q_I_property(self,q_I):
        pI = pifobj.Property()
        n_qpoints = q_I.shape[0]
        pI.scalars = [pifobj.Scalar(q_I[i,1]) for i in range(n_qpoints)]
        pI.units = 'counts'
        pI.conditions = []
        pI.conditions.append( pifobj.Value('scattering vector', 
                            [pifobj.Scalar(q_I[i,0]) for i in range(n_qpoints)],
                            None,None,None,'1/Angstrom') )
        return pI 
        
